ScriptName AproposActorAlias Extends ReferenceAlias

Import AproposUtil
Import SlaveTats

Actor Property ActorRef Auto
Actor Property PlayerRef  Auto
AproposConfig Property config Auto
AproposDescriptionDb Property Database Auto
AproposDescriptions Property Descriptions Auto
SexLabFramework Property SexLab Auto

Int _lastVaginalSexPass = 0
Int _lastAnalSexPass = 0
Int _lastOralSexPass = 0

; raw w&t values.
Int _vaginalWearTear = 0
Int _analWearTear = 0
Int _oralWearTear = 0

; 0 - 9 state variables - indices into W&T effect labels.
Int _vaginalState = 0
Int _analState = 0
Int _oralState = 0

; raw w&t general abuse values
Int _abuseVaginalWearTear = 0
Int _abuseAnalWearTear = 0
Int _abuseOralWearTear = 0

; 0 - 9 state variables - indices into W&T Textures
Int _abuseVaginalState = 0
Int _abuseAnalState = 0
Int _abuseOralState = 0

; raw w&t Creature abuse values
Int _creatureAbuseVaginalWearTear = 0
Int _creatureAbuseAnalWearTear = 0
Int _creatureAbuseOralWearTear = 0

; 0 - 9 state variables - indices into W&T Textures
Int _creatureAbuseVaginalState = 0
Int _creatureAbuseAnalState = 0
Int _creatureAbuseOralState = 0

; raw w&t Daedric abuse values
Int _daedricAbuseVaginalWearTear = 0
Int _daedricAbuseAnalWearTear = 0
Int _daedricAbuseOralWearTear = 0

; 0 - 9 state variables - indices into W&T Textures
Int _daedricAbuseVaginalState = 0
Int _daedricAbuseAnalState = 0
Int _daedricAbuseOralState = 0

Int _aproposId = -1

String _deferredWTChange = ""
String _deferredWTChangeEffectiveVoice = ""
String _deferredWTChangeBodyPart = ""

Int _lastAfterEffectsBodyTxrMapId
Int _lastAfterEffectsFaceTxrMapId
Int _lastCutsScratchesTxrMapId
Int _lastDaedricScarsTxrMapId
Int _lastTearsSobsTxrMapId

Int Function GetId()
    Return _aproposId
EndFunction

Int Function setId(Int newid)
    _aproposId = newid
EndFunction

Int Property AverageAbuseState
    Int Function get()
        Return Average3(_abuseVaginalState, _abuseAnalState, _abuseOralState)
    EndFunction
EndProperty

Int Property AverageCreatureAbuseState
    Int Function get()
        Return Average2(_creatureAbuseVaginalState, _creatureAbuseAnalState)
    EndFunction
EndProperty

Int Property AverageDaedricAbuseState
    Int Function get()
        Return Average2(_daedricAbuseVaginalState, _daedricAbuseAnalState)
    EndFunction
EndProperty

String Property VaginalWearTear
    String Function get()
        Return Database.GetRandomWearAndTearDescriptor(_vaginalState)
    EndFunction
EndProperty

String Property AnalWearTear
    String Function get()
        Return Database.GetRandomWearAndTearDescriptor(_analState)
    EndFunction
EndProperty

String Property OralWearTear
    String Function get()
        Return Database.GetRandomWearAndTearDescriptor(_oralState)
    EndFunction
EndProperty

Int Property VaginalWearTearState
    Int Function get()
        Return _vaginalState
    EndFunction
EndProperty

Int Property AnalWearTearState
    Int Function get()
        Return _analState
    EndFunction
EndProperty

Int Property OralWearTearState
    Int Function get()
        Return _oralState
    EndFunction
EndProperty

Event OnInit()
    GoToState("ReadyForTracking")
EndEvent

; steady state - holding state for unused Alias slot
Auto State ReadyForTracking
    
    Event OnBeginState()
        If config.AreTraceMessagesEnabled()
            Config.Log("AproposActorAlias.ReadyForTracking.OnBeginState")
        EndIf
        ActorRef = none
        Clear()
    EndEvent
    
    Event OnUpdate()
        ; catch any pending updates
    EndEvent

    Function Track(Actor anActor, Int actorSlot)
        If config.AreTraceMessagesEnabled()
            Config.Log("AproposActorAlias.Track(" + anActor.GetDisplayName() + "," + actorSlot + ")")
        EndIf
        ForceRefTo(anActor)
        ActorRef = GetActorRef()
        SetId(actorSlot)
        GoToState("TrackingWearAndTear")
    EndFunction    
EndState

State TrackingWearAndTear
    Event OnBeginState()
        Config.Trace("AproposActorAlias.TrackingWearAndTear.OnBeginState")
        RegisterForSingleUpdateGameTime(1)
    EndEvent
    
    Event OnUpdate()
        If Config.AreTraceMessagesEnabled()
            Config.Log("TrackingWearAndTear.OnUpdate()")
            Config.Log("_deferredWTChange: " + _deferredWTChange + ", _deferredWTChangeEffectiveVoice: " + _deferredWTChangeEffectiveVoice + ", _deferredWTChangeBodyPart: " + _deferredWTChangeBodyPart)
        EndIf
        If _deferredWTChange != "" ; was there a deferred WT change message?
            DisplayWearTearChangedMessage(_deferredWTChange, _deferredWTChangeEffectiveVoice, _deferredWTChangeBodyPart, True)
        EndIf
    EndEvent
    
    Event OnUpdateGameTime()
        UnregisterForUpdateGameTime()

        Int frequency = Config.FrequencyWearTearDegrade
        Float prob = Config.GetReductionProbability(frequency)

        ; Every period (now 1 hour), there is a M probability that reduction will occur.
        If Utility.RandomFloat() <= prob 
            DegradeWearAndTear()
        EndIf
        RegisterForSingleUpdateGameTime(1)
    EndEvent

    Function DisplayWearTearChangedMessage(String change, String effectiveVoice, String bodyPart, Bool scheduled = False)
        If Config.AreTraceMessagesEnabled()
            Config.Log("DisplayWearTearChangedMessage(change=" + change + ", effectiveVoice=" + effectiveVoice + ", bodyPart=" + bodyPart +")")
        EndIf
        Actor anActor = GetActor()
        If IsValidActor(anActor) && !IsActorQuiesced(anActor)
            Int randomWait

            If scheduled
                ; re-schedule
                randomWait = Utility.RandomInt(1, 5) * 60; wait 1 - 5 minutes                
                If config.AreDebugMessagesEnabled()
                    Config.Log("DisplayWearTearChangedMessage: " + anActor.GetDisplayName() + " is in combat or searching or animating. Waiting " + randomWait + "sec.")
                EndIf
                RegisterForSingleUpdate(randomWait)
            ElseIf _deferredWTChange == "" && _deferredWTChangeEffectiveVoice == "" && _deferredWTChangeBodyPart == ""
                ; Only schedule a random update to display the message if we aren't already scheduled!
                _deferredWTChange = change
                _deferredWTChangeEffectiveVoice = effectiveVoice
                _deferredWTChangeBodyPart = bodyPart

                randomWait = Utility.RandomInt(1, 5) * 60; wait 1 - 5 minutes 
                If config.AreDebugMessagesEnabled()
                    Config.Log("DisplayWearTearChangedMessage: " + anActor.GetDisplayName() + " is in combat or searching or animating. Waiting " + randomWait + "sec.")
                EndIf
                RegisterForSingleUpdate(randomWait)
            EndIf

        Else
            If config.AreDebugMessagesEnabled()
                Config.Log("DisplayWearTearChangedMessage: " + anActor.GetDisplayName() + " is not in combat or searching or animating.")
                Config.Log("... Displaying WT changed message...")
            EndIf

            sslBaseVoice voice = SexLab.PickVoice(anActor)
            Int level = Min(AverageAbuseState * 10, 30)
            If change == "Increased"
                Config.WTStaggerSpell.cast(anActor)
                voice.Moan(anActor, level, True)
            Else
                Int arousal = Descriptions.SlaUtil.GetActorArousal(anActor)
                If Config.AutoMasturbateEnabled && (arousal >= Config.MinArousalAutoMasturbate) && IsValidForAutoMasturbate(anActor)

                    String msg = ""
                    If effectiveVoice == "1st Person"
                        msg = "There's an itch ... and I'm so horny..."
                    ElseIf effectiveVoice == "2nd Person"
                        msg = "There's an itch ... and you're so horny..."
                    Else
                        msg = "There's an itch ... and she's so horny..."
                    EndIf

                    If (PlayerRef == anActor)
                        Game.DisablePlayerControls()
                    EndIf
                    Masturbate(anActor)
                    Descriptions.PresentMessage(msg, None)
                    If (PlayerRef == anActor)
                        Game.EnablePlayerControls()
                    EndIf
                Else
                    voice.Moan(anActor, level, False)                    
                EndIf
            EndIf

            Descriptions.DisplayFemaleActorWearTearChangedMessage(change, anActor, effectiveVoice, bodyPart)

            _deferredWTChange = ""
            _deferredWTChangeEffectiveVoice = ""
            _deferredWTChangeBodyPart = ""               
        EndIf
    EndFunction

    Bool Function IsActorQuiesced(Actor anActor)
        Bool isQuiesced = !anActor.IsInCombat() && anActor.GetCombatState() == 0 && !SexLab.IsActorActive(anActor)
        If !isQuiesced && Config.AreDebugMessagesEnabled()
            Config.Log("Actor " + anActor.GetDisplayName() + " is either in combat, searching for combat, or currently animating")
        EndIf
        Return isQuiesced
    EndFunction

    Bool Function IsValidActor(Actor anActor)
        Return anActor && !anActor.IsDead() && !anActor.IsDisabled() && anActor.Is3DLoaded()
    EndFunction

    Bool Function IsValidForAutoMasturbate(Actor anActor)
        Return !anActor.IsOnMount() && !anActor.IsTrespassing() && !anActor.IsGhost()
    EndFunction

    Function Masturbate(Actor anActor)
        Actor[] sexActor = New actor[1]
        sexActor[0] = anActor
        sslBaseAnimation[] anims = SexLab.GetAnimationsByTag(1, "F,Masturbation", TagSuppress="Estrus", RequireAll=True)
        Sexlab.StartSex(sexActor, anims)        
    EndFunction

    ; Int Function MapActorWearValue(Int wearValue, String region, Bool hardcoreMode)
    ;     If hardcoreMode
    ;         region = region + "-hc"
    ;     EndIf

    ;     Int mappedValue = JDB.solveInt(QueryKey("weartearmap", region, wearValue))
    ;     Return mappedValue
    ; EndFunction

    Function UpdateWearTearEffects()

        If !Config.WTEffectsEnabled
            Return
        EndIf

        If Config.AreTraceMessagesEnabled()
            Config.Log("AproposActorAlias.UpdateWearTearEffects for " + GetActor().GetDisplayName(), True)
        EndIf   

        Actor anActor = GetActor()

        Int oralMagickaRegenDebuff
        Int oralSpeechDebuff

        Int vaginalSHRegenDebuff
        Int vaginalMovementDebuff

        Int analSHRegenDebuff
        Int analMovementDebuff

        Bool castOralDebuff = False
        Bool castVaginalDebuff = False
        Bool castAnalDebuff = False

        String level

        DispellAllSpells(anActor)

        If _oralState >= 4

            castOralDebuff = True

            String oral = "oral"

            If config.HardcoreWTEffectsEnabled
                oral += "-hc"
            EndIf             
               
            String keyBase = QueryKey("wearteareffects", oral, "level" + _oralState)

            String oralMagickaRegenDebuffKey = keyBase + ".magickaregen"
            If Config.AreTraceMessagesEnabled()
                Config.Log("oralMagickaRegenDebuffKey = " + oralMagickaRegenDebuffKey)
            EndIf
            oralMagickaRegenDebuff = JDB.solveInt(oralMagickaRegenDebuffKey)

            String oralSpeechDebuffKey = keyBase + ".speech"
            If Config.AreTraceMessagesEnabled()
                Config.Log("oralSpeechDebuffKey = " + oralSpeechDebuffKey)
            EndIf            
            oralSpeechDebuff = JDB.solveInt(oralSpeechDebuffKey)                
        EndIf

        If _vaginalState >= 4

            castVaginalDebuff = True

            String vaginal = "vaginal"

            If config.HardcoreWTEffectsEnabled
                vaginal += "-hc"
            EndIf             
               
            String keyBase = QueryKey("wearteareffects", vaginal, "level" + _vaginalState)

            String vaginalSHRegenDebuffKey = keyBase + ".shregen"
            If Config.AreTraceMessagesEnabled()
                Config.Log("vaginalSHRegenDebuffKey = " + vaginalSHRegenDebuffKey)
            EndIf
            vaginalSHRegenDebuff = JDB.solveInt(vaginalSHRegenDebuffKey)

            String vaginalMovementDebuffKey = keyBase + ".movement"
            If Config.AreTraceMessagesEnabled()
                Config.Log("vaginalMovementDebuffKey = " + vaginalMovementDebuffKey)
            EndIf            
            vaginalMovementDebuff = JDB.solveInt(vaginalMovementDebuffKey)                
        EndIf    

        If _analState >= 4

            castAnalDebuff = True

            String anal = "anal"

            If config.HardcoreWTEffectsEnabled
                anal += "-hc"
            EndIf             
               
            String keyBase = QueryKey("wearteareffects", anal, "level" + _analState)

            String analSHRegenDebuffKey = keyBase + ".shregen"
            Config.Trace("analSHRegenDebuffKey = " + analSHRegenDebuffKey)
            analSHRegenDebuff = JDB.solveInt(analSHRegenDebuffKey)

            String analMovementDebuffKey = keyBase + ".movement"
            If Config.AreTraceMessagesEnabled()
                Config.Log("analMovementDebuffKey = " + analMovementDebuffKey)
            EndIf            
            analMovementDebuff = JDB.solveInt(analMovementDebuffKey)                
        EndIf

        If castOralDebuff
            If oralMagickaRegenDebuff > 0
                Config.OralMagickaDebuffSpell.SetNthEffectMagnitude(0, oralMagickaRegenDebuff)
                anActor.AddSpell(Config.OralMagickaDebuffSpell, False)
                Config.Log("Added oral WT Magicka Debuff spell to actor with magnitude=" + oralMagickaRegenDebuff, True)
            EndIf

            If oralSpeechDebuff > 0
                Config.OralSpeechDebuffSpell.SetNthEffectMagnitude(0, oralSpeechDebuff)
                anActor.AddSpell(Config.OralSpeechDebuffSpell, False)
                Config.Log("Added oral WT Speech Debuff spell to actor with magnitude=" + oralSpeechDebuff, True)
            EndIf
        EndIf

        If castVaginalDebuff
            If vaginalSHRegenDebuff > 0
                Config.VaginalHealthDebuffSpell.SetNthEffectMagnitude(0, vaginalSHRegenDebuff)
                Config.VaginalStaminaDebuffSpell.SetNthEffectMagnitude(0, vaginalSHRegenDebuff)
                anActor.AddSpell(Config.VaginalHealthDebuffSpell, False)
                anActor.AddSpell(Config.VaginalStaminaDebuffSpell, False)
                Config.Log("Added vaginal WT S&H Debuff spell to actor with magnitude=" + vaginalSHRegenDebuff, True)
            EndIf

            If vaginalMovementDebuff > 0
                Config.VaginalSpeedDebuffSpell.SetNthEffectMagnitude(0, vaginalMovementDebuff)
                anActor.AddSpell(Config.VaginalSpeedDebuffSpell, False)
                Config.Log("Added vaginal WT Speed Debuff spell to actor with magnitude=" + vaginalMovementDebuff, True)
            EndIf
        EndIf

        If castAnalDebuff
            If analSHRegenDebuff > 0
                Config.AnalHealthDebuffSpell.SetNthEffectMagnitude(0, analSHRegenDebuff)
                Config.AnalStaminaDebuffSpell.SetNthEffectMagnitude(0, analSHRegenDebuff)
                anActor.AddSpell(Config.AnalHealthDebuffSpell, False)
                anActor.AddSpell(Config.AnalStaminaDebuffSpell, False)
                Config.Log("Added anal WT S&H Debuff spell to actor with magnitude=" + analSHRegenDebuff, True)
            EndIf
            
            If analMovementDebuff > 0
                Config.AnalSpeedDebuffSpell.SetNthEffectMagnitude(0, analMovementDebuff)
                anActor.AddSpell(Config.AnalSpeedDebuffSpell, False)
                Config.Log("Added anal WT Speed Debuff spell to actor with magnitude=" + analMovementDebuff, True)
            EndIf   
        EndIf

    EndFunction    

;Int Function AddTattoo(Actor anActor, String name, String texture, String section, String area, Int slot)

    Function UpdateAbuseTextures()
        If !Config.EnableSkinTextures
            Return
        EndIf

        Actor anActor = GetActor()
        ; remove previous textures
        RemoveAllAbuseTextures(anActor) ; even if its not enabled, remove, in case MCM was changed

        If Config.EnableAfterEffects || Config.EnableTearsAndSobs ; general abuse
            Int avg = Max4(AverageAbuseState, _abuseAnalState, _abuseOralState, _abuseVaginalState)

            Int slot = -1; 1

            If Config.EnableTearsAndSobs

                If avg == 9
                    _lastAfterEffectsFaceTxrMapId = AddTattoo(anActor, "battered", "After Effects\\face_battered.dds", "After Effects", "Face", slot)
                ElseIf avg == 8
                    _lastTearsSobsTxrMapId = AddTattoo(anActor, "Sob 2", "Sob Tears\\sob_2.dds", "Tears", "Face", slot)
                ElseIf avg == 7
                    _lastTearsSobsTxrMapId = AddTattoo(anACtor, "Sob 1", "Sob Tears\\sob_1.dds", "Tears", "Face", slot)
                ElseIf avg == 6
                    _lastTearsSobsTxrMapId = AddTattoo(anACtor, "Tears 3", "Sob Tears\\tears_3.dds", "Tears", "Face", slot)
                ElseIf avg == 5
                    _lastTearsSobsTxrMapId = AddTattoo(anACtor, "Tears 2", "Sob Tears\\tears_2.dds", "Tears", "Face", slot)
                ElseIf avg == 4
                    _lastTearsSobsTxrMapId = AddTattoo(anACtor, "Tears 1", "Sob Tears\\tears_1.dds", "Tears", "Face", slot)
                EndIf
            EndIf            
            
            If Config.EnableAfterEffects
                If avg == 9
                    _lastAfterEffectsBodyTxrMapId = AddTattoo(anActor, "After Effects 4", "After Effects\\after_effects_4.dds", "After Effects", "Body", slot)
                ElseIf avg >= 7
                    _lastAfterEffectsBodyTxrMapId = AddTattoo(anActor, "After Effects 3", "After Effects\\after_effects_3.dds", "After Effects", "Body", slot)
                ElseIf avg >= 6
                    _lastAfterEffectsBodyTxrMapId = AddTattoo(anActor, "After Effects 2", "After Effects\\after_effects_2.dds", "After Effects", "Body", slot)
                ElseIf avg >= 5
                    _lastAfterEffectsBodyTxrMapId = AddTattoo(anActor, "After Effects 1", "After Effects\\after_effects_1.dds", "After Effects", "Body", slot)
                EndIf
            EndIf

        EndIf

        If Config.EnableCutScratches ; creature abuse
            Int avg = Max4(AverageCreatureAbuseState, _creatureAbuseAnalState, _creatureAbuseOralState, _creatureAbuseVaginalState)
            Int slot = -1; 2
            If avg == 9
                _lastCutsScratchesTxrMapId = AddTattoo(anActor, "Cuts 3", "After Effects\\cuts_3.dds", "Cuts", "Body", slot)
            ElseIf avg >= 7
                _lastCutsScratchesTxrMapId = AddTattoo(anActor, "Cuts 2", "After Effects\\cuts_2.dds", "Cuts", "Body", slot)
            ElseIf avg >= 5
                _lastCutsScratchesTxrMapId = AddTattoo(anActor, "Cuts 1", "After Effects\\cuts_1.dds", "Cuts", "Body", slot) 
            EndIf
        EndIf

        If Config.EnableDaedricScars
            Int avg = Max4(AverageDaedricAbuseState, _daedricAbuseAnalState, _daedricAbuseOralState, _daedricAbuseVaginalState)
            Int slot = -1; 3
            If avg == 9
                _lastDaedricScarsTxrMapId = AddTattoo(anActor, "Scars 3", "Scars Dirt\\scarred_3.dds", "Scars", "Body", slot)
            ElseIf avg >= 7
                _lastDaedricScarsTxrMapId = AddTattoo(anActor, "Scars 2", "Scars Dirt\\scarred_2.dds", "Scars", "Body", slot)
            ElseIf avg >= 5
                _lastDaedricScarsTxrMapId = AddTattoo(anActor, "Scars 1", "Scars Dirt\\scarred_1.dds", "Scars", "Body", slot)
            EndIf
        EndIf

        If synchronize_tattoos(anActor, True)
            Config.Debug("Failed to synchronize_tattoos")
        EndIf

    EndFunction

    Function IncreaseVaginalAbuseWearAndTear(Int amountApplied, Bool isRapeOrAggressive, Bool isDaedricCreatureOrDremora, Bool isCreature)
        If isRapeOrAggressive && Config.EnableSkinTextures

            Config.Trace("AproposActorAlias.IncreaseVaginalAbuseWearAndTear("+amountApplied+","+isRapeOrAggressive+","+isDaedricCreatureOrDremora+","+isCreature+")")
                     
            Int oldAbuseVaginalState = _abuseVaginalState

            _abuseVaginalWearTear = CalculateIncreasedWearTearValue(amountApplied, _abuseVaginalWearTear)
            _abuseVaginalState = Config.MapWearTearAmountToState(_abuseVaginalWearTear)
            LogWearTearChange("abuse-vaginal", _abuseVaginalWearTear, _abuseVaginalState)

            If isCreature && Config.EnableCutScratches

                Int oldCreatureAbuseVaginalState = _creatureAbuseVaginalState
                _creatureAbuseVaginalWearTear = CalculateIncreasedWearTearValue(amountApplied, _creatureAbuseVaginalWearTear)
                _creatureAbuseVaginalState = Config.MapWearTearAmountToState(_creatureAbuseVaginalWearTear)
                LogWearTearChange("creature-abuse-vaginal", _creatureAbuseVaginalWearTear, _creatureAbuseVaginalState)

            EndIf

            If isDaedricCreatureOrDremora && Config.EnableDaedricScars

                Int oldDaedricAbuseVaginalState = _daedricAbuseVaginalState
                _daedricAbuseVaginalWearTear = CalculateIncreasedWearTearValue(amountApplied, _daedricAbuseVaginalWearTear)
                _daedricAbuseVaginalState = Config.MapWearTearAmountToState(_daedricAbuseVaginalWearTear)
                LogWearTearChange("daedric-abuse-vaginal", _daedricAbuseVaginalWearTear, _daedricAbuseVaginalState)                

            EndIf
        EndIf

    EndFunction

    Function IncreaseVaginalWearAndTear(Int amountApplied)
        If !GuardWearTearEnabled("IncreaseVaginalWearAndTear", amountApplied)
            Return
        EndIf

        Int oldWearAndTearState = _vaginalState
        
        _vaginalWearTear = CalculateIncreasedWearTearValue(amountApplied, _vaginalWearTear)
        _vaginalState = Config.MapWearTearAmountToState(_vaginalWearTear)
        
        LogWearTearChange("vaginal", _vaginalWearTear, _vaginalState)

        Int change = _vaginalState - oldWearAndTearState

        If change > 0 && _vaginalState > 0
            String effectiveVoice = Descriptions.GetEffectiveNarrativeVoice(GetActor() == PlayerRef)
            DisplayWearTearChangedMessage("Increased", effectiveVoice, "Vaginal")
        EndIf

        _lastVaginalSexPass = 0
    EndFunction

    Function IncreaseAnalAbuseWearAndTear(Int amountApplied, Bool isRapeOrAggressive, Bool isDaedricCreatureOrDremora, Bool isCreature)
        If isRapeOrAggressive && Config.EnableSkinTextures

            Config.Trace("AproposActorAlias.IncreaseAnalAbuseWearAndTear("+amountApplied+","+isRapeOrAggressive+","+isDaedricCreatureOrDremora+","+isCreature+")")

            Int oldAbuseAnalState = _abuseAnalState

            _abuseAnalWearTear = CalculateIncreasedWearTearValue(amountApplied, _abuseAnalWearTear)
            _abuseAnalState = Config.MapWearTearAmountToState(_abuseAnalWearTear)
            LogWearTearChange("abuse-anal", _abuseAnalWearTear, _abuseAnalState)

            If isCreature && Config.EnableCutScratches

                Int oldCreatureAbuseAnalState = _creatureAbuseAnalState
                _creatureAbuseAnalWearTear = CalculateIncreasedWearTearValue(amountApplied, _creatureAbuseAnalWearTear)
                _creatureAbuseAnalState = Config.MapWearTearAmountToState(_creatureAbuseAnalWearTear)
                LogWearTearChange("creature-abuse-anal", _creatureAbuseAnalWearTear, _creatureAbuseAnalState)

            EndIf

            If isDaedricCreatureOrDremora && Config.EnableDaedricScars

                Int oldDaedricAbuseAnalState = _daedricAbuseAnalState
                _daedricAbuseAnalWearTear = CalculateIncreasedWearTearValue(amountApplied, _daedricAbuseAnalWearTear)
                _daedricAbuseAnalState = Config.MapWearTearAmountToState(_daedricAbuseAnalWearTear)
                LogWearTearChange("daedric-abuse-anal", _daedricAbuseAnalWearTear, _daedricAbuseAnalState)                
            EndIf
        EndIf                
    EndFunction

    Function IncreaseAnalWearAndTear(Int amountApplied)
        If !GuardWearTearEnabled("ApplyIncreasedAnalWearAndTearAmount", amountApplied)
            Return
        EndIf

        Int oldWearAndTearState = _analState
        
        _analWearTear = CalculateIncreasedWearTearValue(amountApplied, _analWearTear)
        _analState = Config.MapWearTearAmountToState(_analWearTear)
        
        LogWearTearChange("anal", _analWearTear, _analState)

        Int change = _analState - oldWearAndTearState

        If change > 0 && _analState > 0
            String effectiveVoice = Descriptions.GetEffectiveNarrativeVoice(GetActor() == PlayerRef)
            DisplayWearTearChangedMessage("Increased", effectiveVoice, "Anal")
        EndIf            
        
        _lastAnalSexPass = 0
    EndFunction

    Function IncreaseOralAbuseWearAndTear(Int amountApplied, Bool isRapeOrAggressive, Bool isDaedricCreatureOrDremora, Bool isCreature)
        If isRapeOrAggressive && Config.EnableSkinTextures

            Config.Trace("AproposActorAlias.IncreaseOralAbuseWearAndTear("+amountApplied+","+isRapeOrAggressive+","+isDaedricCreatureOrDremora+","+isCreature+")")

            Int oldAbuseOralState = _abuseOralState

            _abuseOralWearTear = CalculateIncreasedWearTearValue(amountApplied, _abuseOralWearTear)
            _abuseOralState = Config.MapWearTearAmountToState(_abuseOralWearTear)
            LogWearTearChange("abuse-oral", _abuseOralWearTear, _abuseOralState)

            If isCreature && Config.EnableCutScratches

                Int oldCreatureAbuseOralState = _creatureAbuseOralState
                _creatureAbuseOralWearTear = CalculateIncreasedWearTearValue(amountApplied, _creatureAbuseOralWearTear)
                _creatureAbuseOralState = Config.MapWearTearAmountToState(_creatureAbuseOralWearTear)
                LogWearTearChange("creature-abuse-oral", _creatureAbuseOralWearTear, _creatureAbuseOralState)

            EndIf

            If isDaedricCreatureOrDremora && Config.EnableDaedricScars

                Int oldDaedricAbuseOralState = _daedricAbuseOralState
                _daedricAbuseOralWearTear = CalculateIncreasedWearTearValue(amountApplied, _daedricAbuseOralWearTear)
                _daedricAbuseOralState = Config.MapWearTearAmountToState(_daedricAbuseOralWearTear)
                LogWearTearChange("daedric-abuse-oral", _daedricAbuseOralWearTear, _daedricAbuseOralState)                

            EndIf
        EndIf                
    EndFunction

    Function IncreaseOralWearAndTear(Int amountApplied)
        If !GuardWearTearEnabled("ApplyIncreasedOralWearAndTearAmount", amountApplied)
            Return
        EndIf

        Int oldWearAndTearState = _oralState
        
        _oralWearTear = CalculateIncreasedWearTearValue(amountApplied, _oralWearTear)
        _oralState = Config.MapWearTearAmountToState(_oralWearTear)
        
        LogWearTearChange("oral", _oralWearTear, _oralState)

        Int change = _oralState - oldWearAndTearState

        If change > 0 && _oralState > 0
            String effectiveVoice = Descriptions.GetEffectiveNarrativeVoice(GetActor() == PlayerRef)
            DisplayWearTearChangedMessage("Increased", effectiveVoice, "Oral")
        EndIf 
        
        _lastOralSexPass = 0
    EndFunction

    Function ApplyReducedVaginalWearAndTearAmount(Int amountApplied)
        If !GuardWearTearEnabled("ApplyReducedVaginalWearAndTearAmount", amountApplied)
            Return
        EndIf

        Int oldWearAndTearState = _vaginalState
        
        _vaginalWearTear = CalculateReducedWearTearValue(amountApplied, _vaginalWearTear)
        _vaginalState = Config.MapWearTearAmountToState(_vaginalWearTear)
        LogWearTearChange("vaginal", _vaginalWearTear, _vaginalState)

        Int change = _vaginalState - oldWearAndTearState

        If change < 0 && _vaginalState > 0
            String effectiveVoice = Descriptions.GetEffectiveNarrativeVoice(GetActor() == PlayerRef)
            DisplayWearTearChangedMessage("Reduced", effectiveVoice, "Vaginal")
        EndIf
    EndFunction

    Function ApplyReducedVaginalAbuseWearAndTearAmount(Int amountApplied)
        If Config.EnableSkinTextures

            Config.Trace("AproposActorAlias.ApplyReducedVaginalWearAndTearAmount("+amountApplied+")")
                     
            Int oldAbuseVaginalState = _abuseVaginalState

            _abuseVaginalWearTear = CalculateReducedWearTearValue(amountApplied, _abuseVaginalWearTear)
            _abuseVaginalState = Config.MapWearTearAmountToState(_abuseVaginalWearTear)
            LogWearTearChange("abuse-vaginal", _abuseVaginalWearTear, _abuseVaginalState)

            If Config.EnableCutScratches

                Int oldCreatureAbuseVaginalState = _creatureAbuseVaginalState
                _creatureAbuseVaginalWearTear = CalculateReducedWearTearValue(amountApplied, _creatureAbuseVaginalWearTear)
                _creatureAbuseVaginalState = Config.MapWearTearAmountToState(_creatureAbuseVaginalWearTear)
                LogWearTearChange("creature-abuse-vaginal", _creatureAbuseVaginalWearTear, _creatureAbuseVaginalState)

            ElseIf Config.EnableDaedricScars

                Int oldDaedricAbuseVaginalState = _daedricAbuseVaginalState
                _daedricAbuseVaginalWearTear = CalculateReducedWearTearValue(amountApplied, _daedricAbuseVaginalWearTear)
                _daedricAbuseVaginalState = Config.MapWearTearAmountToState(_daedricAbuseVaginalWearTear)
                LogWearTearChange("daedric-abuse-vaginal", _daedricAbuseVaginalWearTear, _daedricAbuseVaginalState)                

            EndIf
        EndIf
    EndFunction

    Function ApplyReducedAnalWearAndTearAmount(Int amountApplied)
        If !GuardWearTearEnabled("ApplyReducedAnalWearAndTearAmount", amountApplied)
            Return
        EndIf

        Int oldWearAndTearState = _analState
        
        _analWearTear = CalculateReducedWearTearValue(amountApplied, _analWearTear)
        _analState = Config.MapWearTearAmountToState(_analWearTear)
        LogWearTearChange("anal", _analWearTear, _analState)  

        Int change = _analState - oldWearAndTearState

        If change < 0 && _analState > 0
            String effectiveVoice = Descriptions.GetEffectiveNarrativeVoice(GetActor() == PlayerRef)
            DisplayWearTearChangedMessage("Reduced", effectiveVoice, "Anal")
        EndIf        
    EndFunction

    Function ApplyReducedAnalAbuseWearAndTearAmount(Int amountApplied)
        If Config.EnableSkinTextures

            Config.Trace("AproposActorAlias.ApplyReducedAnalAbuseWearAndTearAmount("+amountApplied+")")

            Int oldAbuseAnalState = _abuseAnalState

            _abuseAnalWearTear = CalculateReducedWearTearValue(amountApplied, _abuseAnalWearTear)
            _abuseAnalState = Config.MapWearTearAmountToState(_abuseAnalWearTear)
            LogWearTearChange("abuse-anal", _abuseAnalWearTear, _abuseAnalState)

            If Config.EnableCutScratches

                Int oldCreatureAbuseAnalState = _creatureAbuseAnalState
                _creatureAbuseAnalWearTear = CalculateReducedWearTearValue(amountApplied, _creatureAbuseAnalWearTear)
                _creatureAbuseAnalState = Config.MapWearTearAmountToState(_creatureAbuseAnalWearTear)
                LogWearTearChange("creature-abuse-anal", _creatureAbuseAnalWearTear, _creatureAbuseAnalState)

            ElseIf Config.EnableDaedricScars

                Int oldDaedricAbuseAnalState = _daedricAbuseAnalState
                _daedricAbuseAnalWearTear = CalculateReducedWearTearValue(amountApplied, _daedricAbuseAnalWearTear)
                _daedricAbuseAnalState = Config.MapWearTearAmountToState(_daedricAbuseAnalWearTear)
                LogWearTearChange("daedric-abuse-anal", _daedricAbuseAnalWearTear, _daedricAbuseAnalState)                
            EndIf
        EndIf                
    EndFunction    

    Function ApplyReducedOralWearAndTearAmount(Int amountApplied)
        If !GuardWearTearEnabled("ApplyReducedOralWearAndTearAmount", amountApplied)
            Return
        EndIf

        Int oldWearAndTearState = _oralState

        _oralWearTear = CalculateReducedWearTearValue(amountApplied, _oralWearTear)
        _oralState = Config.MapWearTearAmountToState(_oralWearTear)
        LogWearTearChange("oral", _oralWearTear, _oralState)  

        Int change = _oralState - oldWearAndTearState

        If change < 0 && _oralState > 0
            String effectiveVoice = Descriptions.GetEffectiveNarrativeVoice(GetActor() == PlayerRef)
            DisplayWearTearChangedMessage("Reduced", effectiveVoice, "Oral")
        EndIf         
    EndFunction        

    Function ApplyReducedOralAbuseWearAndTearAmount(Int amountApplied)
        If Config.EnableSkinTextures

            Config.Trace("AproposActorAlias.ApplyReducedOralAbuseWearAndTearAmount("+amountApplied+")")

            Int oldAbuseOralState = _abuseOralState

            _abuseOralWearTear = CalculateReducedWearTearValue(amountApplied, _abuseOralWearTear)
            _abuseOralState = Config.MapWearTearAmountToState(_abuseOralWearTear)
            LogWearTearChange("abuse-oral", _abuseOralWearTear, _abuseOralState)

            If Config.EnableCutScratches

                Int oldCreatureAbuseOralState = _creatureAbuseOralState
                _creatureAbuseOralWearTear = CalculateReducedWearTearValue(amountApplied, _creatureAbuseOralWearTear)
                _creatureAbuseOralState = Config.MapWearTearAmountToState(_creatureAbuseOralWearTear)
                LogWearTearChange("creature-abuse-oral", _creatureAbuseOralWearTear, _creatureAbuseOralState)

            ElseIf Config.EnableDaedricScars

                Int oldDaedricAbuseOralState = _daedricAbuseOralState
                _daedricAbuseOralWearTear = CalculateReducedWearTearValue(amountApplied, _daedricAbuseOralWearTear)
                _daedricAbuseOralState = Config.MapWearTearAmountToState(_daedricAbuseOralWearTear)
                LogWearTearChange("daedric-abuse-oral", _daedricAbuseOralWearTear, _daedricAbuseOralState)                

            EndIf
        EndIf                
    EndFunction     
    
    Function DegradeWearAndTear()
        String name = GetActor().GetDisplayName()

        If Config.AreDebugMessagesEnabled()
            Config.Log("Reducing " + name + "'s sexual wear and tear...", toConsole=True)
        EndIf

        If (Config.AreDebugMessagesEnabled())
            Config.Log("AproposActorAlias.DegradeWearAndTear")
            Config.Log(name + ": Last Vaginal Sex Pass : " + _lastVaginalSexPass)
            Config.Log(name + ": Last Anal Sex Pass : " + _lastAnalSexPass)
            Config.Log(name + ": Last Oral Sex Pass : " + _lastOralSexPass)
        EndIf
        
        Float degradeFactor = config.WTDegradeFactor
        
        Int vaginalWTReduction =  (Math.pow(_lastVaginalSexPass, degradeFactor) * 15.0) As Int
        Int analWTReduction =  (Math.pow(_lastAnalSexPass, degradeFactor) * 15.0) As Int
        Int oralWTReduction =  (Math.pow(_lastOralSexPass, degradeFactor) * 15.0) As Int
        
        If (config.AreDebugMessagesEnabled())
            Config.Log("WTDegradeFactor : " + degradeFactor)
            Config.Log(name + ": Vaginal WT reduction amount : " + vaginalWTReduction)
            Config.Log(name + ": Anal WT reduction amount : " + analWTReduction)
            Config.Log(name + ": Oral WT reduction amount : " + oralWTReduction)
        EndIf 
        
        ApplyReducedVaginalWearAndTearAmount(vaginalWTReduction)
        ApplyReducedAnalWearAndTearAmount(analWTReduction)
        ApplyReducedOralWearAndTearAmount(oralWTReduction)

        ApplyReducedVaginalAbuseWearAndTearAmount(vaginalWTReduction)
        ApplyReducedAnalAbuseWearAndTearAmount(analWTReduction)
        ApplyReducedOralAbuseWearAndTearAmount(oralWTReduction * 2); double the reduction on oral-abuse so face textures change faster

        LogAllWearAndTear()
        UpdateWearTearEffects()
        UpdateAbuseTextures()
        
        _lastVaginalSexPass += 1
        _lastAnalSexPass += 1
        _lastOralSexPass += 1

        If (_vaginalWearTear <= 1 && _analWearTear <= 1 && _oralWearTear <= 1)
            GoToState("ClearTracking")
        EndIf
    EndFunction

    Function LogAllWearAndTear()
        If !Config.AreTraceMessagesEnabled()
            Return
        EndIf
        Config.Log("Vaginal=" + _vaginalState + "(" + _vaginalWearTear +")", True)
        Config.Log("- AbuseVaginal=" + _abuseVaginalState + "(" + _abuseVaginalWearTear + ")", True)
        Config.Log("- CreatureAbuseVaginal=" + _creatureAbuseVaginalState + "(" + _creatureAbuseVaginalWearTear + ")", True)
        Config.Log("- DaedricAbuseVaginal=" + _daedricAbuseVaginalState + "(" + _daedricAbuseVaginalWearTear + ")", True)

        Config.Log("Anal=" + AnalWearTearState + "(" + AnalWearTear +")", True)
        Config.Log("- AbuseAnal=" + _abuseAnalState + "(" + _abuseAnalWearTear + ")", True)
        Config.Log("- CreatureAbuseAnal=" + _creatureAbuseAnalState + "(" + _creatureAbuseAnalWearTear + ")", True)
        Config.Log("- DaedricAbuseAnal=" + _daedricAbuseAnalState + "(" + _daedricAbuseAnalWearTear + ")", True)

        Config.Log("Oral=" + OralWearTearState + "(" + OralWearTear +")", True)
        Config.Log("- AbuseOral=" + _abuseOralState + "(" + _abuseOralWearTear + ")", True)
        Config.Log("- CreatureAbuseOral=" + _creatureAbuseOralState + "(" + _creatureAbuseOralWearTear + ")", True)
        Config.Log("- DaedricAbuseOral=" + _daedricAbuseOralState + "(" + _daedricAbuseOralWearTear + ")", True)
        Config.Log("- avg Abuse = " + AverageAbuseState, True)
        Config.Log("- avg Creature Abuse = " + AverageCreatureAbuseState, True)
        Config.Log("- avg Daedric Abuse = " + AverageDaedricAbuseState, True)
    EndFunction
    
    Event OnEndState()
        
    EndEvent		
    
    Event OnDeath(Actor akKiller)
        GoToState("ReadyForTracking")
    EndEvent

    Function OverrideVaginalWT(Int newValue, Bool updateAbuse, Bool updateCreatureAbuse, Bool updateDaedricAbuse)
        _vaginalState = newValue
        ; make sure we update the wear tear amount inferred from state.
        Int wearTearAmount = Config.MapWearTearStateToAmount(newValue)
        _vaginalWearTear = wearTearAmount
        If Config.EnableSkinTextures
            If updateAbuse 
                _abuseVaginalState = newValue
                _abuseVaginalWearTear = wearTearAmount
            EndIf

            If updateCreatureAbuse
                _creatureAbuseVaginalState = newValue
                _creatureAbuseVaginalWearTear = wearTearAmount
            EndIf

            If updateDaedricAbuse
                _daedricAbuseVaginalState = newValue
                _daedricAbuseVaginalWearTear = wearTearAmount
            EndIf
        EndIf
    EndFunction

    Function OverrideAnalWT(Int newValue, Bool updateAbuse, Bool updateCreatureAbuse, Bool updateDaedricAbuse)
        _analState = newValue
        ; make sure we update the wear tear amount inferred from state.
        Int wearTearAmount = Config.MapWearTearStateToAmount(newValue)
        _analWearTear = wearTearAmount
        If Config.EnableSkinTextures
            If updateAbuse 
                _abuseAnalState = newValue
                _abuseAnalWearTear = wearTearAmount
            EndIf

            If updateCreatureAbuse
                _creatureAbuseAnalState = newValue
                _creatureAbuseAnalWearTear = wearTearAmount
            EndIf

            If updateDaedricAbuse
                _daedricAbuseAnalState = newValue
                _daedricAbuseAnalWearTear = wearTearAmount
            EndIf
        EndIf        
    EndFunction

    Function OverrideOralWT(Int newValue, Bool updateAbuse, Bool updateCreatureAbuse, Bool updateDaedricAbuse)
        _oralState = newValue
        ; make sure we update the wear tear amount inferred from state.
        Int wearTearAmount = Config.MapWearTearStateToAmount(newValue)
        _oralWearTear = wearTearAmount
        If Config.EnableSkinTextures
            If updateAbuse 
                _abuseOralState = newValue
                _abuseOralWearTear = wearTearAmount
            EndIf

            If updateCreatureAbuse
                _creatureAbuseOralState = newValue
                _creatureAbuseOralWearTear = wearTearAmount
            EndIf

            If updateDaedricAbuse
                _daedricAbuseOralState = newValue
                _daedricAbuseOralWearTear = wearTearAmount
            EndIf
        EndIf         
    EndFunction
EndState

State ClearTracking
    Event OnBeginState()
        Config.Trace("AproposActorAlias.ClearTracking.OnBeginState")
        UnregisterForUpdate()

        Actor anActor = GetActor()

        If anActor
            RemoveAllAbuseTextures(anActor, True)
            DispellAllSpells(anActor)
        EndIf

        _oralState = 0
        _oralWearTear = 0
        _lastOralSexPass = 0

        _analState = 0
        _analWearTear = 0
        _lastAnalSexPass = 0

        _vaginalState = 0
        _vaginalWearTear = 0
        _lastVaginalSexPass = 0    

        _abuseVaginalWearTear = 0
        _abuseAnalWearTear = 0
        _abuseOralWearTear = 0

        _abuseVaginalState = 0
        _abuseAnalState = 0
        _abuseOralState = 0

        _creatureAbuseVaginalWearTear = 0
        _creatureAbuseAnalWearTear = 0
        _creatureAbuseOralWearTear = 0

        _creatureAbuseVaginalState = 0
        _creatureAbuseAnalState = 0
        _creatureAbuseOralState = 0

        _daedricAbuseVaginalWearTear = 0
        _daedricAbuseAnalWearTear = 0
        _daedricAbuseOralWearTear = 0

        _daedricAbuseVaginalState = 0
        _daedricAbuseAnalState = 0
        _daedricAbuseOralState = 0        

        _deferredWTChange = ""
        _deferredWTChangeEffectiveVoice = ""
        _deferredWTChangeBodyPart = ""        

        GoToState("ReadyForTracking")
    EndEvent
    
    Event OnEndState()
        
    EndEvent
    
    Event OnUpdate()
        
    EndEvent
EndState

Function DegradeWearAndTear()
    ; Do nothing if not in TrackingWearAndTear state
EndFunction

Actor Function GetActor()
    Return ActorRef
EndFunction

Function LogWearTearChange(String type, Int wearTearValue, Int wearTearState)
    If config.AreDebugMessagesEnabled()
        String name = ActorRef.GetDisplayName()
        Config.Log(name + ": " + type + " Wear & Tear = " + wearTearValue + " (value) " + wearTearState + " (state)")
    EndIf  
EndFunction

Bool Function GuardWearTearEnabled(String label, Int amount)
    If amount == 0
        Return False
    EndIf

    If !config.IsWearAndTearEnabled
        Return False
    EndIf
    
    If ActorRef != PlayerRef && !config.IsNPCWearAndTearEnabled
        Config.Debug("NPC W&T disabled and this is not the player")
        Return False
    EndIf    
    
    If (config.AreTraceMessagesEnabled())
        String actorName = ActorRef.GetDisplayName()
        Config.Log("AproposActorAlias." + label + "(" + amount + ") for actor " + actorName)
    EndIf    
    
    Return True
EndFunction

Int Function CalculateIncreasedWearTearValue(Int amountApplied, Int wearTearAmount)
    If Config.AreTraceMessagesEnabled()
        Config.Log("AproposActorAlias.CalculateIncreasedWearTearValue(" + amountApplied + "," + wearTearAmount + ")")
    EndIf
    Int range = amountApplied / 5
    Int randomAmount = Utility.RandomInt(-range, range) + amountApplied
    
    Int res = SslUtility.ClampInt(wearTearAmount + randomAmount, 0, Config.MaxWearTearAmount)
    Config.Trace("AproposActorAlias.CalculateIncreasedWearTearValue final calcuation: " + res)
    Return res
EndFunction

Int Function CalculateReducedWearTearValue(Int amountApplied, Int wearTearAmount)
    Int range = amountApplied / 5
    Int randomAmount = Utility.RandomInt(-range, range) + amountApplied
    If amountApplied > wearTearAmount
        wearTearAmount = 1
    ElseIf amountApplied < wearTearAmount
        wearTearAmount -= randomAmount
    EndIf
    Return SslUtility.ClampInt(wearTearAmount, 0, Config.MaxWearTearAmount)
EndFunction

Function RemoveAllAbuseTextures(Actor anActor, Bool synchronize = False)
    RemoveAfterEffectsTexture(anActor)
    RemoveTearsSobsTexture(anActor)
    RemoveCutsScratchesTexture(anActor)
    RemoveDaedricScarsTexture(anActor)
    If synchronize && synchronize_tattoos(anActor, True)
        Config.Debug("Failed to synchronize_tattoos")
    EndIf
EndFunction

Function RemoveAfterEffectsTexture(Actor anActor)
    Config.Trace("AproposActorAlias.RemoveAfterEffectsTexture", True)
    If _lastAfterEffectsBodyTxrMapId != 0
        If remove_tattoos(anActor, _lastAfterEffectsBodyTxrMapId)
            Config.Debug("Failed to remove tattoo - AfterEffectsBody")
        EndIf
        JValue.release(_lastAfterEffectsBodyTxrMapId)
        _lastAfterEffectsBodyTxrMapId = 0
    EndIf
EndFunction

Function RemoveTearsSobsTexture(Actor anActor)
    Config.Trace("AproposActorAlias.RemoveTearsSobsTexture", True)
    If _lastTearsSobsTxrMapId != 0
        If remove_tattoos(anActor, _lastTearsSobsTxrMapId)
            Config.Debug("Failed to remove tattoo - TearsSobs")
        EndIf
        JValue.release(_lastTearsSobsTxrMapId)
        _lastTearsSobsTxrMapId = 0        
    EndIf
    If _lastAfterEffectsFaceTxrMapId != 0
        If remove_tattoos(anActor, _lastAfterEffectsFaceTxrMapId)
            Config.Debug("Failed to remove tattoo - AfterEffectsFace")
        EndIf
        JValue.release(_lastAfterEffectsFaceTxrMapId)
        _lastAfterEffectsFaceTxrMapId = 0        
    EndIf    
EndFunction

Function RemoveCutsScratchesTexture(Actor anActor)
    Config.Trace("AproposActorAlias.RemoveCutsScratchesTexture", True)
    If _lastCutsScratchesTxrMapId != 0
        If remove_tattoos(anActor, _lastCutsScratchesTxrMapId)
            Config.Debug("Failed to remove tattoo - CutsScratches")
        EndIf
        JValue.release(_lastCutsScratchesTxrMapId)
        _lastCutsScratchesTxrMapId = 0             
    EndIf 
EndFunction

Function RemoveDaedricScarsTexture(Actor anActor)
    Config.Trace("AproposActorAlias.RemoveDaedricScarsTexture", True)
    If _lastDaedricScarsTxrMapId != 0
        If remove_tattoos(anActor, _lastDaedricScarsTxrMapId)
            Config.Debug("Failed to remove tattoo - DaedricScars")
        EndIf
        JValue.release(_lastDaedricScarsTxrMapId)
        _lastDaedricScarsTxrMapId = 0            
    EndIf    
EndFunction

Function DispellAllSpells(Actor anActor)
    Config.Debug("Removing debuff spells from " + anACtor.GetDisplayName())

    If anActor.HasSpell(Config.OralMagickaDebuffSpell)
        anActor.RemoveSpell(Config.OralMagickaDebuffSpell)
    EndIf
    If anActor.HasSpell(Config.OralSpeechDebuffSpell)
        anActor.RemoveSpell(Config.OralSpeechDebuffSpell)
    EndIf
    If anActor.HasSpell(Config.AnalHealthDebuffSpell)
        anActor.RemoveSpell(Config.AnalHealthDebuffSpell)
    EndIf
    If anActor.HasSpell(Config.AnalStaminaDebuffSpell)
        anActor.RemoveSpell(Config.AnalStaminaDebuffSpell)
    EndIf    
    If anActor.HasSpell(Config.AnalSpeedDebuffSpell)
        anActor.RemoveSpell(Config.AnalSpeedDebuffSpell)
    EndIf
    If anActor.HasSpell(Config.VaginalHealthDebuffSpell)
        anActor.RemoveSpell(Config.VaginalHealthDebuffSpell)
    EndIf
    If anActor.HasSpell(Config.VaginalStaminaDebuffSpell)
        anActor.RemoveSpell(Config.VaginalStaminaDebuffSpell)
    EndIf    
    If anActor.HasSpell(Config.VaginalSpeedDebuffSpell)
        anActor.RemoveSpell(Config.VaginalSpeedDebuffSpell)
    EndIf                    
EndFunction

Int Function AddTattoo(Actor anActor, String name, String texture, String section, String area, Int slot)
    ; AddTattoo(anActor, "Sob 2", "Sob Tears\\sob_2.dds", "Tears", "Face", slot)
    Int colorTint = config.TatsColorTint
    Int map = RetainedMap4Pair("name", name, "texture", texture, "section", section, "area", area)
    JMap.setInt(map, "color", colorTint)
    If Config.AreTraceMessagesEnabled()
        Config.Log("AproposActorAlias.AddTattoo, calling add_tattoo", True)
        Config.Log("... section : " + section, True)
        Config.Log("... area : " + area, True)
        Config.Log("... name : " + name, True)
        Config.Log("... texture : " + texture, True)
        Config.Log("... slot : " + slot, True)
        Config.Log("... color: " + colorTint, True)
        Config.Log("... actor : " + anActor.GetDisplayName(), True)
    EndIf    
    If add_tattoo(anActor, map, slot)
        Config.Debug("Failed to add tattoo")
    EndIf
    Return map
EndFunction

Bool Function IsActorQuiesced(Actor anActor)
    Config.Log("Attempts to call IsActorQuiesced while in empty state. Should be called in TrackingWearAndTear state")
    Return False
    ; Specifically do nothing in the default state    
EndFunction

Bool Function IsValidForAutoMasturbate(Actor anActor)
    Config.Log("Attempts to call IsValidForAutoMasturbate while in empty state. Should be called in TrackingWearAndTear state")
    Return False
    ; Specifically do nothing in the default state   
EndFunction

Bool Function IsValidActor(Actor anActor)
    Config.Log("Attempts to call IsValidActor while in empty state. Should be called in TrackingWearAndTear state")
    Return False
    ; Specifically do nothing in the default state       
EndFunction

Function Masturbate(Actor anActor)
    Config.Log("Attempts to call Masturbate while in empty state. Should be called in TrackingWearAndTear state")
    ; Specifically do nothing in the default state           
EndFunction

Function ApplyReducedVaginalWearAndTearAmount(Int amountApplied)
    Config.Log("Attempts to call ApplyReducedVaginalWearAndTearAmount while in empty state. Should be called in TrackingWearAndTear state")
    ; Specifically do nothing in the default state
EndFunction

Function ApplyReducedVaginalAbuseWearAndTearAmount(Int amountApplied)
    Config.Log("Attempts to call ApplyReducedVaginalAbuseWearAndTearAmount while in empty state. Should be called in TrackingWearAndTear state")
    ; Specifically do nothing in the default state
EndFunction

Function ApplyReducedAnalWearAndTearAmount(Int amountApplied)
    Config.Log("Attempts to call ApplyReducedAnalWearAndTearAmount while in empty state. Should be called in TrackingWearAndTear state")
    ; Specifically do nothing in the default state
EndFunction

Function ApplyReducedAnalAbuseWearAndTearAmount(Int amountApplied)
    Config.Log("Attempts to call ApplyReducedAnalAbuseWearAndTearAmount while in empty state. Should be called in TrackingWearAndTear state")
    ; Specifically do nothing in the default state
EndFunction    

Function ApplyReducedOralWearAndTearAmount(Int amountApplied)
    Config.Log("Attempts to call ApplyReducedOralWearAndTearAmount while in empty state. Should be called in TrackingWearAndTear state")
    ; Specifically do nothing in the default state
EndFunction

Function ApplyReducedOralAbuseWearAndTearAmount(Int amountApplied)
    Config.Log("Attempts to call ApplyReducedOralAbuseWearAndTearAmount while in empty state. Should be called in TrackingWearAndTear state")
    ; Specifically do nothing in the default state
EndFunction

Function IncreaseVaginalWearAndTear(Int amountApplied)
    Config.Log("Attempts to call IncreaseVaginalWearAndTear while in empty state. Should be called in TrackingWearAndTear state")
    ; Specifically do nothing in the default state
EndFunction

Function IncreaseVaginalAbuseWearAndTear(Int amountApplied, Bool isRapeOrAggressive, Bool isDaedricCreatureOrDremora, Bool isCreature)
    Config.Log("Attempts to call IncreaseVaginalAbuseWearAndTear while in empty state. Should be called in TrackingWearAndTear state")
    ; Specifically do nothing in the default state
EndFunction

Function IncreaseAnalWearAndTear(Int amountApplied)
    Config.Log("Attempts to call IncreaseAnalWearAndTear while in empty state. Should be called in TrackingWearAndTear state")
    ; Specifically do nothing in the default state
EndFunction

Function IncreaseAnalAbuseWearAndTear(Int amountApplied, Bool isRapeOrAggressive, Bool isDaedricCreatureOrDremora, Bool isCreature)
    Config.Log("Attempts to call IncreaseAnalAbuseWearAndTear while in empty state. Should be called in TrackingWearAndTear state")
    ; Specifically do nothing in the default state
EndFunction

Function IncreaseOralWearAndTear(Int amountApplied)
    Config.Log("Attempts to call IncreaseOralWearAndTear while in empty state. Should be called in TrackingWearAndTear state")
    ; Specifically do nothing in the default state
EndFunction

Function IncreaseOralAbuseWearAndTear(Int amountApplied, Bool isRapeOrAggressive, Bool isDaedricCreatureOrDremora, Bool isCreature)
    Config.Log("Attempts to call IncreaseOralAbuseWearAndTear while in empty state. Should be called in TrackingWearAndTear state")
    ; Specifically do nothing in the default state
EndFunction

Function Track(Actor anActor, Int actorSlot)
    Config.Log("Attempts to call Track while in empty state. Should be called in ReadyForTracking state")
    ; Specifically do nothing in the default state
EndFunction

Function DisplayWearTearChangedMessage(String change, String effectiveVoice, String bodyPart, Bool scheduled = False)
    Config.Log("Attempts to call DisplayWearTearChangedMessage while in empty state. Should be called in ReadyForTracking state")
    ; Specifically do nothing in the default state    
EndFunction

Function UpdateWearTearEffects()
    Config.Log("Attempts to call ApplyWearTearEffects while in empty state. Should be called in ReadyForTracking state")
    ; Specifically do nothing in the default state  
EndFunction

String Function TestIsAlive()
    Return "Alive"
EndFunction

Function UpdateAbuseTextures()
    Config.Log("Attempts to call ApplyAbuseTextures while in empty state. Should be called in ReadyForTracking state")
    ; Specifically do nothing in the default state  
EndFunction

Function LogAllWearAndTear()
    Config.Log("Attempts to call LogAllWearAndTear while in empty state. Should be called in ReadyForTracking state")
    ; Specifically do nothing in the default state     
EndFunction

Function OverrideVaginalWT(Int newValue, Bool updateAbuse, Bool updateCreatureAbuse, Bool updateDaedricAbuse)
    Config.Log("Attempts to call OverrideVaginalWT while in empty state. Should be called in ReadyForTracking state")
    ; Specifically do nothing in the default state  
EndFunction

Function OverrideAnalWT(Int newValue, Bool updateAbuse, Bool updateCreatureAbuse, Bool updateDaedricAbuse)
    Config.Log("Attempts to call OverrideVaginalWT while in empty state. Should be called in ReadyForTracking state")
    ; Specifically do nothing in the default state  
EndFunction

Function OverrideOralWT(Int newValue, Bool updateAbuse, Bool updateCreatureAbuse, Bool updateDaedricAbuse)
    Config.Log("Attempts to call OverrideVaginalWT while in empty state. Should be called in ReadyForTracking state")
    ; Specifically do nothing in the default state  
EndFunction