ScriptName AproposConfig Extends SKI_ConfigBase

Actor Property PlayerRef Auto ; //Hook Player
AproposDescriptionDb Property Database Auto
Apropos Property MainQuest Auto
AproposAnimationPatchups Property AnimationPatchups Auto
AproposCommon Property Common Auto
SexLabFramework Property SexLab  Auto
AproposWidgetManager Property Widgets Auto
AproposDescriptions Property Descriptions Auto

Spell Property AnalHealthDebuffSpell Auto
Spell Property AnalStaminaDebuffSpell Auto
Spell Property AnalSpeedDebuffSpell Auto
Spell Property OralMagickaDebuffSpell Auto
Spell Property OralSpeechDebuffSpell Auto
Spell Property VaginalHealthDebuffSpell Auto
Spell Property VaginalStaminaDebuffSpell Auto
Spell Property VaginalSpeedDebuffSpell Auto
Spell Property WTStaggerSpell Auto

Int Property AproposVersion Auto

Int _requiredJCAPIVersion = 3
Int _minimumJCfeatureVersion = 0

Int flag
Int slvflag

String[] _personOptions
String[] _horizontalAnchors
String[] _verticalAnchors
String _slaveTatsStatus

Import AproposUtil

Int Function GetVersion()
    Return 20170109
EndFunction

Event OnVersionUpdate(Int newVersion)
    Log("AproposConfig.OnVersionUpdate(" + newVersion + ")")
    If newVersion >= 20170109 && CurrentVersion < 20170109
        Log("Upgrading Apropos to: " + GetVersion())
        Debug.Notification("Upgrading Apropos to: " + GetVersion())
        OnConfigInit()
        UpgradeMainQuest()
    EndIf
EndEvent

Event OnGameReload()
    parent.OnGameReload()

    If ValidateJContainers() != "Valid"
        Return
    EndIf   

    InitPages()    

    MainQuest.GameLoaded()
    Widgets.GameLoaded()
EndEvent

Function InitPages()

    If ValidateJContainers() != "Valid"
        Pages = New String[1]
        Pages[0] = "Derp"
        Return
    EndIf
    
    Pages = New String[7]
    Pages[0] = "General"
    Pages[1] = "Wear and Tear Settings"
    Pages[2] = "Wear and Tear Actors"
    Pages[3] = "Message Preferences"
    Pages[4] = "Events and Messages"
    Pages[5] = "Message Widget Settings"
    Pages[6] = "Message Themes"
EndFunction

Event OnConfigInit()
    ModName = "Apropos"

    Log("AproposConfig.OnConfigInit")

    InitPages()
    
    _personOptions = New String[3]
    _personOptions[0] = "1st Person"
    _personOptions[1] = "2nd Person"
    _personOptions[2] = "3rd Person"

    _horizontalAnchors = New String[3]
    _horizontalAnchors[0] = "Left"
    _horizontalAnchors[1] = "Center"
    _horizontalAnchors[2] = "Right"
    _verticalAnchors = New String[3]
    _verticalAnchors[0] = "Top"
    _verticalAnchors[1] = "Center"
    _verticalAnchors[2] = "Bottom"

    _personIndex = 1
    UpdateNarrativeMode()

    _uninstalling = False
    _runningPatchups = False
    _importing = False
    _exporting = False

EndEvent

Event OnConfigClose()
    If (_uninstalling)
        Debug.Notification("All Apropos quests were stopped.")
        Debug.Notification("Save your game and remove Apropos from your Skyrim installation.")
    Else
        ; Widgets.UpdateWidgets()
    EndIf
EndEvent

Function UpgradeMainQuest()
    Database.Initialize()
EndFunction

Function UpdateNarrativeMode()
    NarrativeMode = _personOptions[_personIndex]
EndFunction

Event OnConfigOpen()
    
    If PlayerRef.GetActorBase().GetSex() == 0
        flag = OPTION_FLAG_DISABLED
        
    ElseIf PlayerRef.GetActorBase().GetSex() == 1
        flag = OPTION_FLAG_NONE
    EndIf
    
EndEvent

Bool Property IsJCValid
    Bool Function get()
        Return ValidateJContainers() == "Valid"
    EndFunction
EndProperty

Int _clearNpcWearTearOID

Int _vaginalStateIndex
Int _oralStateIndex
Int _analStateIndex

Int _vaginalStateOID
Int _analStateOID
Int _oralStateOID

Int _avgGeneralAbuseOID
Int _avgCreatureAbuseOID
Int _avgDaedricAbuseOID

Int _horizontalAnchorOID
Int _verticalAnchorOID
Int _horizontalOffsetOID
Int _verticalOffsetOID
Int _fontSizeOID
Int _playerFontSizeOID
Int _widgetSpacingOID
Int _secondsToDisplayOID

Int _testButtonOID

Int _minArousalToAutoMasturbateOID
Int Property MinArousalAutoMasturbate = 80 Auto Hidden

Int _AutoMasturbateEnabledOID
Bool Property AutoMasturbateEnabled = True Auto Hidden

Int _tatsColorTintOID
; See for codes: https://github.com/schlangster/skyui/blob/master/src/Common/skyui/components/colorswatch/ColorSwatch.as
Int Property TatsColorTint = 0x990033 Auto Hidden ; 0x990033 is deep red

Int _wtDegradeFactorOID
Float _wtDegradeFactor = 1.2
Float Property WTDegradeFactor
    Float Function get()
        Return _wtDegradeFactor
    EndFunction       
EndProperty

String Property NarrativeMode Auto Hidden
Int _personMenuOID
Int _personIndex

Int _enableSkinTexturesOID
Bool _enableSkinTextures = False
Bool Property EnableSkinTextures
    Bool Function get()
        Return _enableSkinTextures
    EndFunction
EndProperty

Int _enableAfterEffectsOID
Bool _enableAfterEffects = True
Bool Property EnableAfterEffects
    Bool Function get()
        Return _enableAfterEffects
    EndFunction    
EndProperty

Int _enableCutScratchesOID
Bool _enableCutScratches = True
Bool Property EnableCutScratches
    Bool Function get()
        Return _enableCutScratches
    EndFunction    
EndProperty

Int _enableDaedricScarsOID
Bool _enableDaedricScars = True
Bool Property EnableDaedricScars
    Bool Function get()
        Return _enableDaedricScars
    EndFunction    
EndProperty

Int _enableTearsAndSobsOID
Bool _enableTearsAndSobs = True
Bool Property EnableTearsAndSobs
    Bool Function get()
        Return _enableTearsAndSobs
    EndFunction    
EndProperty

Int _enableConsoleOutputOID
Bool _enableConsoleOutput = True
Bool Property EnableConsoleOutput
    Bool Function get()
        Return _enableConsoleOutput
    EndFunction  
EndProperty

Int _showWTChangedMessagesOID
Bool _showWTChangedMessages = True
Bool Property ShowWTChangedMessages
    Bool Function get()
        Return _showWTChangedMessages
    EndFunction
EndProperty

Int _showVirginityLostMessagesOID
Bool _showVirginityLostMessages = True
Bool Property ShowVirginityLostMessages
    Bool Function get()
        Return _showVirginityLostMessages
    EndFunction 
EndProperty

Int _showSexDescriptionsOID
Bool _showSexDescriptions = True
Bool Property ShowSexDescriptions
    Bool Function get()
        Return _showSexDescriptions
    EndFunction    
EndProperty

Int _hardcoreWTEffectsEnabledOID
Bool _hardcoreWTEffectsEnabled = False
Bool Property HardcoreWTEffectsEnabled
    Bool Function get()
        Return _hardcoreWTEffectsEnabled
    EndFunction
EndProperty

Bool _exporting
Bool _importing

Int _exportMCMOID
Int _importMCMOID

Int _widgetTestModeOID

Int _uninstallModOID
Bool _uninstalling

Int _runAnimationPatchupsOID
Bool _runningPatchups

Int _wtEffectsEnabledOID
Bool _wtEffectsEnabled = True
Bool Property WTEffectsEnabled
    Bool Function get()
        Return _wtEffectsEnabled
    EndFunction
EndProperty

Int _logActorStatsOID
Bool _logActorStatsEnabled = False
Bool Property LogActorStatsEnabled
    Bool Function get()
        Return _logActorStatsEnabled
    EndFunction
EndProperty

Int _frequencyWTDegradeOID
Int _frequencyWTDegrade = 12; in hours, default 12 hours
Int Property FrequencyWearTearDegrade
    Int Function get()
        Return _frequencyWTDegrade
    EndFunction
EndProperty

Int _showAnimationStartMessagesOID
Bool _showAnimationStartMessages = False
Bool Property ShowAnimationStartMessages
    Bool Function get()
        return _showAnimationStartMessages
    EndFunction
EndProperty

Int _showAnimationChangeMessagesOID
Bool _showAnimationChangeMessages = False
Bool Property ShowAnimationChangeMessages
    Bool Function get()
        return _showAnimationChangeMessages
    EndFunction
EndProperty

Int _showStage1MessagesOID
Bool _showStage1Messages = True
Bool Property ShowStage1Messages
    Bool Function get()
        Return _showStage1Messages
    EndFunction
EndProperty

Int _showStage2MessagesOID
Bool _showStage2Messages = True
Bool Property ShowStage2Messages
    Bool Function get()
        Return _showStage2Messages
    EndFunction
EndProperty

Int _showStage3MessagesOID
Bool _showStage3Messages = True
Bool Property ShowStage3Messages
    Bool Function get()
        Return _showStage3Messages
    EndFunction
EndProperty

Int _showStage4MessagesOID
Bool _showStage4Messages = True
Bool Property ShowStage4Messages
    Bool Function get()
        Return _showStage4Messages
    EndFunction
EndProperty

Int _showStage5MessagesOID
Bool _showStage5Messages = True
Bool Property ShowStage5Messages
    Bool Function get()
        Return _showStage5Messages
    EndFunction
EndProperty

Int _showStage6MessagesOID
Bool _showStage6Messages = True
Bool Property ShowStage6Messages
    Bool Function get()
        Return _showStage6Messages
    EndFunction
EndProperty

Int _showStage7MessagesOID
Bool _showStage7Messages = True
Bool Property ShowStage7Messages
    Bool Function get()
        Return _showStage7Messages
    EndFunction
EndProperty

Int _showStage8MessagesOID
Bool _showStage8Messages = True
Bool Property ShowStage8Messages
    Bool Function get()
        Return _showStage8Messages
    EndFunction
EndProperty

Int _showStage9MessagesOID
Bool _showStage9Messages = True
Bool Property ShowStage9Messages
    Bool Function get()
        Return _showStage9Messages
    EndFunction
EndProperty

Int _showStage10MessagesOID
Bool _showStage10Messages = True
Bool Property ShowStage10Messages
    Bool Function get()
        Return _showStage10Messages
    EndFunction
EndProperty

Int _showStage11MessagesOID
Bool _showStage11Messages = True
Bool Property ShowStage11Messages
    Bool Function get()
        Return _showStage11Messages
    EndFunction
EndProperty

Int _showStage12MessagesOID
Bool _showStage12Messages = True
Bool Property ShowStage12Messages
    Bool Function get()
        Return _showStage12Messages
    EndFunction
EndProperty

Int _showStage13MessagesOID
Bool _showStage13Messages = True
Bool Property ShowStage13Messages
    Bool Function get()
        Return _showStage13Messages
    EndFunction
EndProperty

Int _enableDebugOID
Bool _debugMessagesEnabled = True
Bool Property DebugMessagesEnabled
    Bool Function get()
        Return _debugMessagesEnabled
    EndFunction
EndProperty

Int _enableTraceOID
Bool _traceMessagesEnabled = False
Bool Property TraceMessagesEnabled
    Bool Function get()
        Return _traceMessagesEnabled
    EndFunction
EndProperty

Int _enableModOID
Bool _isModEnabled = True  
Bool  Property IsModEnabled
    Bool Function get()
        Return _isModEnabled
    EndFunction
EndProperty

Int _enableWearAndTearOID
Bool _isWearAndTearEnabled = True
Bool Property IsWearAndTearEnabled
    Bool Function get()
        Return _isWearAndTearEnabled
    EndFunction
EndProperty

Int _enableNPCWearAndTearOID
Bool _isNPCWearAndTearEnabled = False
Bool Property IsNPCWearAndTearEnabled
    Bool Function get()
        Return _isNPCWearAndTearEnabled
    EndFunction
EndProperty

Int _refreshFromDatabaseOID

Int _maxWearTearAmountOID
Int _maxWearTearAmount = 2000
Int Property MaxWearTearAmount 
    Int Function get()
        Return _maxWearTearAmount
    EndFunction
EndProperty

Int _widgetTestHotkeyOID

Bool Function AreDebugMessagesEnabled()
    Return DebugMessagesEnabled
EndFunction

Bool Function AreTraceMessagesEnabled()
    Return TraceMessagesEnabled
EndFunction

Function CheckForMods()
    Bool slaveTatsOn = False

    Int i
    While i < Game.GetModCount()
        If Game.GetModName(i) == "Slavetats.esp"
            slaveTatsOn = True
            slvflag = OPTION_FLAG_NONE
            Log("Found Slavetats", True)
            _slaveTatsStatus = "Found Slavetats"
        EndIf
        i += 1
    EndWhile

    If !slaveTatsOn
        Log("Did not find Slavetats", True)
        slvflag = OPTION_FLAG_DISABLED
        _slaveTatsStatus = "Did not find Slavetats"
        _enableSkinTextures = False
    EndIf

EndFunction

Event OnPageReset(String page)

    If page == ""
        LoadCustomContent("Apropos/Apropos.dds",125 ,35.5)
        Return
    EndIf

    UnloadCustomContent()
    SetCursorFillMode(TOP_TO_BOTTOM) ; by default all pages top->bottom

    If _uninstalling
        AddHeaderOption("UNINSTALLING")
        AddEmptyOption()
        AddHeaderOption("Complete the uninstallation")
        AddHeaderOption("or reload a previous save.")

    ElseIf _runningPatchups
        AddHeaderOption("Animation Patchups Running")

    ElseIf _exporting
        AddHeaderOption("Exporting Settings")
    ElseIf _importing
        AddHeaderOption("Importing Settings")

    ElseIf page == "General"
        AddHeaderOption("Apropos Version : " + GetVersion())
        AddEmptyOption()

        _enableModOID = AddToggleOption("Enable mod", IsModEnabled)
        _enableDebugOID = AddToggleOption("Enable debug messages", DebugMessagesEnabled)
        _enableTraceOID = AddToggleOption("Enable detailed trace messages", TraceMessagesEnabled)
        _logActorStatsOID = AddToggleOption("Enable Actor SexLab Stats Logging", LogActorStatsEnabled)
        _enableConsoleOutputOID = AddToggleOption("Enable output to Skyrim Console", EnableConsoleOutput)
        AddEmptyOption()
        _runAnimationPatchupsOID = AddTextOption("SexLab Animation Patchups", "RUN PATCHUPS")
        AddEmptyOption()
        _uninstallModOID = AddTextOption("Uninstall mod", "UNINSTALL")

        SetCursorPosition(1)
        AddHeaderOption("Export and Import Settings")
        _exportMCMOID = AddTextOption("Export Apropos Settings", "EXPORT")
        _importMCMOID = AddTextOption("Import Apropos Settings", "IMPORT")

    ElseIf page == "Wear and Tear Settings"
        _enableWearAndTearOID = AddToggleOption("Enable Apropos Wear and Tear", IsWearAndTearEnabled)
        _enableNPCWearAndTearOID = AddToggleOption("Enable NPC Apropos Wear and Tear", IsNPCWearAndTearEnabled)
        _wtDegradeFactorOID = AddSliderOption("Wear and Tear Degrade Factor", WTDegradeFactor, "{1}")
        _maxWearTearAmountOID = AddSliderOption("Maximum Wear Tear Value", MaxWearTearAmount)
        _frequencyWTDegradeOID = AddSliderOption("Frequency of Wear Tear Degrade", FrequencyWearTearDegrade)
        AddHeaderOption("Wear Tear Effects/Consequences")
        _wtEffectsEnabledOID = AddToggleOption("Enable Wear Tear Effects", WTEffectsEnabled)
        _hardcoreWTEffectsEnabledOID = AddToggleOption("Enable Hardcore Wear Tear Effects", HardcoreWTEffectsEnabled)
        _autoMasturbateEnabledOID = AddToggleOption("Enable Auto-Masturbate", AutoMasturbateEnabled)
        _minArousalToAutoMasturbateOID = AddSliderOption("Minimum Arousal for Auto-Masturbate", MinArousalAutoMasturbate)
        _testButtonOID= AddTextOption("Test W&T on Player", "TEST")
        SetCursorPosition(1)
        AddHeaderOption("Textures for W&T 'Abuse' - Requires Slavetats")

        CheckForMods()

        AddTextOption("Slavetats Mod status", _slaveTatsStatus)
        _enableSkinTexturesOID     = AddToggleOption("Enable SkinTextures for Wear and Tear", EnableSkinTextures, slvflag)
        _enableAfterEffectsOID     = AddToggleOption("Enable 'After Effects' Textures", EnableAfterEffects, slvflag)
        _enableCutScratchesOID     = AddToggleOption("Enable 'Cuts and Scratches' Textures", EnableCutScratches, slvflag)
        _enableDaedricScarsOID     = AddToggleOption("Enable Daedric 'Scars' Textures", EnableDaedricScars, slvflag)
        _enableTearsAndSobsOID     = AddToggleOption("Enable 'Tears & Sobs' Textures", EnableTearsAndSobs, slvflag)

        _tatsColorTintOID          = AddColorOption("Tattoo Color Tine", TatsColorTint)

    ElseIf page == "Message Preferences"
        AddHeaderOption("Narrative Mode (Grammatical Person)")
        _personMenuOID = AddMenuOption("Person", _personOptions[_personIndex])
        AddEmptyOption()
        _refreshFromDatabaseOID = AddTextOption("Refresh From Database", "REFRESH")
        AddEmptyOption()
        _showSexDescriptionsOID = AddToggleOption("Show SexLab Animation Descriptions", ShowSexDescriptions)
        _showVirginityLostMessagesOID = AddToggleOption("Show Virginity Lost Messages", ShowVirginityLostMessages)
        _showWTChangedMessagesOID = AddToggleOption("Show Wear & Tear Changed Messages", ShowWTChangedMessages)

    ElseIf page == "Wear and Tear Actors"
        AddHeaderOption("Current Player Wear and Tear Values")
        AproposActorAlias playerAlias = MainQuest.GetAproposActor(PlayerRef)
        If !playerAlias
            AddTextOption("Player - no W&T abuse yet (not tracked).", "", OPTION_FLAG_DISABLED)
        Else
            _vaginalStateOID = AddMenuOption("Vaginal State", Database.GetMCMWearAndTearDescriptor(playerAlias.VaginalWearTearState), flag)
            _analStateOID = AddMenuOption("Anal State", Database.GetMCMWearAndTearDescriptor(playerAlias.AnalWearTearState))
            _oralStateOID = AddMenuOption("Oral State", Database.GetMCMWearAndTearDescriptor(playerAlias.OralWearTearState))
            If EnableSkinTextures
                AddHeaderOption("Wear and Tear Abuse Statistics")
                _avgGeneralAbuseOID = AddTextOption("General Avg Abuse: ", Database.GetMCMWearAndTearDescriptor(playerAlias.AverageAbuseState))
                _avgCreatureAbuseOID = AddTextOption("Creature Avg Abuse: ", Database.GetMCMWearAndTearDescriptor(playerAlias.AverageCreatureAbuseState))
                _avgDaedricAbuseOID = AddTextOption("Daedric Avg Abuse: ", Database.GetMCMWearAndTearDescriptor(playerAlias.AverageDaedricAbuseState))
            EndIf
        EndIf   
        SetCursorPosition(1)
        AddHeaderOption("Current NPC Wear Tear Actors")

        If !IsNPCWearAndTearEnabled
            AddTextOption("Wear and Tear is currently disabled for NPCs", "", OPTION_FLAG_DISABLED)
        Else
            String[] alist = MainQuest.GetTrackedActorList()

            Bool anyTrackedNpcs = False

            Int i = 0 
            While i < alist.Length
                If alist[i] != ""
                    AddTextOption(alist[i], "")
                    anyTrackedNpcs = True
                EndIf
                i += 1
            EndWhile
            
            If anyTrackedNpcs
                _clearNpcWearTearOID = AddTextOption("Clear NPC Wear Tear Tracking", "CLEAR")
            Else
                AddTextOption("No NPCs currently have their wear and tear tracked.", "", OPTION_FLAG_DISABLED)
            EndIf
        EndIf

    ElseIf page == "Events and Messages"
        _showAnimationStartMessagesOID = AddToggleOption("Enable Animation Start Messages", ShowAnimationStartMessages)
        _showAnimationChangeMessagesOID = AddToggleOption("Enable Animation Change Messages", ShowAnimationChangeMessages)
        _showStage1MessagesOID = AddToggleOption("Enable Stage 1 Progression Messages", ShowStage1Messages)
        _showStage2MessagesOID = AddToggleOption("Enable Stage 2 Progression Messages", ShowStage2Messages)
        _showStage3MessagesOID = AddToggleOption("Enable Stage 3 Progression Messages", ShowStage3Messages)
        _showStage4MessagesOID = AddToggleOption("Enable Stage 4 Progression Messages", ShowStage4Messages)
        _showStage5MessagesOID = AddToggleOption("Enable Stage 5 Progression Messages", ShowStage5Messages)
        _showStage6MessagesOID = AddToggleOption("Enable Stage 6 Progression Messages", ShowStage6Messages)
        _showStage7MessagesOID = AddToggleOption("Enable Stage 7 Progression Messages", ShowStage7Messages)
        _showStage8MessagesOID = AddToggleOption("Enable Stage 8 Progression Messages", ShowStage8Messages)
        _showStage9MessagesOID = AddToggleOption("Enable Stage 9 Progression Messages", ShowStage9Messages)
        _showStage10MessagesOID = AddToggleOption("Enable Stage 10 Progression Messages", ShowStage10Messages)
        _showStage11MessagesOID = AddToggleOption("Enable Stage 11 Progression Messages", ShowStage11Messages)
        _showStage12MessagesOID = AddToggleOption("Enable Stage 12 Progression Messages", ShowStage12Messages)
        _showStage13MessagesOID = AddToggleOption("Enable Stage 13 Progression Messages", ShowStage13Messages)

    ElseIf page == "Message Widget Settings"

        AddHeaderOption("Widget Position")
        _horizontalAnchorOID = AddMenuOption("Horizontal anchor", Widgets.HorizontalAnchor)
        _horizontalOffsetOID = AddSliderOption("Horizontal offset", Widgets.HorizontalOffset, "{0} pixels")
        _verticalAnchorOID = AddMenuOption("Vertical anchor", Widgets.VerticalAnchor)
        _verticalOffsetOID = AddSliderOption("Vertical offset", Widgets.VerticalOffset, "{0} pixels")

        AddHeaderOption("Widget Appearance")
        _fontSizeOID = AddSliderOption("Font Size", Widgets.FontSize, "{0} pixels")
        _playerFontSizeOID = AddSliderOption("Player Font Size", Widgets.PlayerFontSize, "{0} pixels")
        _widgetSpacingOID = AddSliderOption("Widget Spacing", Widgets.WidgetSpacing, "{0} pixels")
        _secondsToDisplayOID = AddSliderOption("Seconds To Display", Widgets.SecondsToDisplay, "{1} secs")

        SetCursorPosition(1)

        _widgetTestModeOID = AddToggleOption("Widget Test Mode", Widgets.TestMode)
        _widgetTestHotkeyOID = AddKeyMapOption("Widget Test Hotkey", MainQuest.WidgetTestHotkey, OPTION_FLAG_WITH_UNMAP)

    ElseIf page == "Derp"
        AddHeaderOption(ValidateJContainers())

    ElseIf page == "Message Themes"
        AddHeaderOption("Message Themes")
        Int mapId = JDB.solveObj(QueryKey("themes-raw"))
        String[] allKeys = StringArrayFromJMapKeys(mapId)
        Int i = 0
        While i < allKeys.Length
            Int val = JDB.solveInt(QueryKey("themes-raw", allKeys[i], "weight"))
            AddTextOption(allKeys[i], (val As String))
            i += 1
        EndWhile
        
    EndIf
    
EndEvent

; Highlighting (affects Text, Menus, ToggleOptions, Sliders)

Event OnOptionHighlight(Int option)
    If _uninstalling
        SetInfoText("UNINSTALLING")
        Return
    EndIf

    If _runningPatchups
        SetInfoText("Running patchups")
        Return
    EndIf

    If _importing
        SetInfoText("Importing settings")
        Return
    EndIf

    If _exporting
        SetInfoText("Exporting settings")
        Return
    EndIf

    If option == _personMenuOID
        SetInfoText("1st person:'The Bandit cums in my ass'\n2nd Person:'The Bandit cums in your ass'\n3rd Person:'The Bandit cums in Amanda's ass' (Amanda is the PC).")
    ElseIf option == _enableDebugOID
        SetInfoText("Emit SexLab animation event information into the Apropos User Logs.")
    ElseIf option == _enableTraceOID
        SetInfoText("Emit detailed trace messages into the Apropos User Logs.")
    ElseIf option == _vaginalStateOID
        SetInfoText("Override the current state of the player's vaginal wear and tear.")
    ElseIf option == _analStateOID
        SetInfoText("Override the current state of the player's anal wear and tear.")
    ElseIf option == _oralStateOID
        SetInfoText("Override the current state of the player's oral wear and tear.")
    ElseIf option == _refreshFromDatabaseOID
        SetInfoText("If you have customized any of the Json CONTROL .txt files, use this to refresh your game. Example: Synonyms.txt. This is not required for description file changes.")
    ElseIf option == _enableWearAndTearOID
        SetInfoText("Uncheck to disable the Apropos Wear and Tear system.")
    ElseIf option == _enableNPCWearAndTearOID
        SetInfoText("Uncheck to disable the Apropos Wear and Tear system NPCs including Followers.")
    ElseIf option == _wtDegradeFactorOID
        SetInfoText("Determines how Wear and Tear exponentially degrades. Set to 1.0 for linear progression.")
    ElseIf option == _clearNpcWearTearOID        
        SetInfoText("Clear all NPC Wear & Tear tracking.")
    ElseIf option == _maxWearTearAmountOID
        SetInfoText("Maximum raw Wear Tear value.")
    ElseIf option == _logActorStatsOID
        SetInfoText("Enable logging of stats (Pain, Enjoyment, Lewdness) to Apropos.log and Skyrim Console.")
    ElseIf option == _frequencyWTDegradeOID
        SetInfoText("How frequent, in hours, the Wear Tear factors are degraded.")
    ElseIf option == _uninstallModOID
        SetInfoText("Stops main quest, removes JContainers data.")
    ElseIf option == _runAnimationPatchupsOID
        SetInfoText("Runs SexLab Animation Patchups script which adjusts various tags.")
    ElseIf option == _showSexDescriptionsOID
        SetInfoText("Unselect to suppress all animation start, stage start, and orgasm descriptions. Wear & Tear messages are still enabled.")
    ElseIf option == _enableConsoleOutputOID
        SetInfoText("Unselect to suppress all Apropos console window output.")
    ElseIf option == _showVirginityLostMessagesOID
        SetInfoText("Unselect to suppress all messages about virginity being lost.")
    ElseIf option == _showWTChangedMessagesOID
        SetInfoText("Unselect to suppress all messages regarding Wear and Tear changes.")
    ElseIf option == _enableSkinTexturesOID
        SetInfoText("Select to enable SkinTextures to be applied to an actor (PC or NPC) that corresponds to the level of W&T 'abuse' the character has sustained. Consensual sex will not cause application.")
    ElseIf option == _enableAfterEffectsOID
        SetInfoText("Enable to apply general W&T abuse textures crossing different encounters.")
    ElseIf option == _enableCutScratchesOID
        SetInfoText("Enable to apply W&T abuse textures associated with non-Daedric creatures.")
    ElseIf option == _enableDaedricScarsOID
        SetInfoText("Enable to apply W&T abuse textures associated with Daedrice creatures (e.g. Seekers, Lurkers) and Dremora.")
    ElseIf option == _enableTearsAndSobsOID
        SetInfoText("Enable to apply W&T abuse textures related to female crying, etc.")
    ElseIf option == _horizontalAnchorOID
        SetInfoText("Sets the horizontal docking position of the Apropr Widget: Left, Center or Right. Note: for best performance on slow computers, dock the Widget on the Left. Right and Center require extra processing to properly align the text.")
    ElseIf option == _horizontalOffsetOID
        SetInfoText("Sets the horizontal offset of the Apropos Message Widget. Negative values move the Apropos Message Widget to the left, positive values move it to the right.")
    ElseIf option == _verticalAnchorOID
        SetInfoText("Sets the vertical position of the Apropos Message Widget: Top, Center or Bottom.")
    ElseIf option == _verticalOffsetOID
        SetInfoText("Sets the vertical offset of the Apropos Message Widget. Negative values move the Apropos Message Widget up, positive values move it down.")
    ElseIf option == _fontSizeOID
        SetInfoText("Sets the font size used by the Apropos Message Widget.")
    ElseIf option == _widgetSpacingOID
        SetInfoText("Adjust the spacing between the instances of the Apropos Message Widget. Negative values move the widgets closer together, positive values move them further apart.")
    ElseIf option == _exportMCMOID
        SetInfoText("Export settings under " + GenerateSettingsExportPath())
    ElseIf option == _importMCMOID
        SetInfoText("Import settings under " + GenerateSettingsExportPath()) 
    ElseIf option == _widgetTestModeOID
        SetInfoText("Enables Widgets test mode.")
    ElseIf option == _widgetTestHotkeyOID
        SetInfoText("Sets the Widgets test hotkey.")
    ElseIf option == _secondsToDisplayOID
        SetInfoText("How many seconds to display messages in each widget instance.")
    ElseIf option == _minArousalToAutoMasturbateOID
        SetInfoText("Minimum arousal to auto-masturbate when W&T has reduced.")
    ElseIf option == _playerFontSizeOID
        SetInfoText("Sets the PC font size used by the Apropos Message Widget.")       
    ElseIf option == _autoMasturbateEnabledOID
        SetInfoText("Enables Auto-Masturbate during W&T healing.") 
    ElseIf option == _testButtonOID
        SetInfoText("Will set the Player to have high levels of W&T.")
    EndIf
EndEvent

; Text & Toggles ...

Event OnOptionSelect(Int option)
    If option == _enableDebugOID
        _debugMessagesEnabled = !_debugMessagesEnabled
        SetToggleOptionValue(_enableDebugOID, DebugMessagesEnabled)
    ElseIf option == _enableTraceOID
        _traceMessagesEnabled = !_traceMessagesEnabled
        SetToggleOptionValue(_enableTraceOID, TraceMessagesEnabled)        
    ElseIf option == _enableModOID
        _isModEnabled = !_isModEnabled
        SetToggleOptionValue(_enableModOID, IsModEnabled)
    ElseIf option == _refreshFromDatabaseOID
        UpgradeMainQuest()
        ForcePageReset()
        ShowMessage("Database Refreshed. Check Console.")
    ElseIf option == _enableWearAndTearOID
        _isWearAndTearEnabled = !_isWearAndTearEnabled
        SetToggleOptionValue(_enableWearAndTearOID, IsWearAndTearEnabled)
        If !IsWearAndTearEnabled 
            MainQuest.ClearPlayerActorTracking()
            ForcePageReset()
            ShowMessage("Player Wear and Tear tracking cleared.")            
        EndIf
    ElseIf option == _enableNPCWearAndTearOID
        _isNPCWearAndTearEnabled = !_isNPCWearAndTearEnabled
        SetToggleOptionValue(_enableNPCWearAndTearOID, IsNPCWearAndTearEnabled)     
        If !IsNPCWearAndTearEnabled
            ClearNPCTrackings()
        EndIf
    ElseIf option == _clearNpcWearTearOID 
        ClearNPCTrackings()
    ElseIf option == _logActorStatsOID
        _logActorStatsEnabled = !_logActorStatsEnabled
        SetToggleOptionValue(_logActorStatsOID, LogActorStatsEnabled)
    ElseIf option == _showStage1MessagesOID
        _showStage1Messages = !_showStage1Messages
        SetToggleOptionValue(_showStage1MessagesOID, ShowStage1Messages)
    ElseIf option == _showStage2MessagesOID
        _showStage2Messages = !_showStage2Messages
        SetToggleOptionValue(_showStage2MessagesOID, ShowStage2Messages)
    ElseIf option == _showStage3MessagesOID
        _showStage3Messages = !_showStage3Messages
        SetToggleOptionValue(_showStage3MessagesOID, ShowStage3Messages)
    ElseIf option == _showStage4MessagesOID
        _showStage4Messages = !_showStage4Messages
        SetToggleOptionValue(_showStage4MessagesOID, ShowStage4Messages)
    ElseIf option == _showStage5MessagesOID
        _showStage5Messages = !_showStage5Messages
        SetToggleOptionValue(_showStage5MessagesOID, ShowStage5Messages)
    ElseIf option == _showStage6MessagesOID
        _showStage6Messages = !_showStage6Messages
        SetToggleOptionValue(_showStage6MessagesOID, ShowStage6Messages)
    ElseIf option == _showStage7MessagesOID
        _showStage7Messages = !_showStage7Messages
        SetToggleOptionValue(_showStage7MessagesOID, ShowStage7Messages)
    ElseIf option == _showStage8MessagesOID
        _showStage8Messages = !_showStage8Messages
        SetToggleOptionValue(_showStage8MessagesOID, ShowStage8Messages)
    ElseIf option == _showStage9MessagesOID
        _showStage9Messages = !_showStage9Messages
        SetToggleOptionValue(_showStage9MessagesOID, ShowStage9Messages)
    ElseIf option == _showStage10MessagesOID
        _showStage10Messages = !_showStage10Messages
        SetToggleOptionValue(_showStage10MessagesOID, ShowStage10Messages)
    ElseIf option == _showStage11MessagesOID
        _showStage11Messages = !_showStage11Messages
        SetToggleOptionValue(_showStage11MessagesOID, ShowStage11Messages)
    ElseIf option == _showStage12MessagesOID
        _showStage12Messages = !_showStage12Messages
        SetToggleOptionValue(_showStage12MessagesOID, ShowStage12Messages)
    ElseIf option == _showStage13MessagesOID
        _showStage13Messages = !_showStage13Messages
        SetToggleOptionValue(_showStage13MessagesOID, ShowStage13Messages)                                        
    ElseIf option == _showAnimationStartMessagesOID
        _showAnimationStartMessages = !_showAnimationStartMessages
        SetToggleOptionValue(_showAnimationStartMessagesOID, ShowAnimationStartMessages)
    ElseIf option == _showAnimationChangeMessagesOID
        _showAnimationChangeMessages = !_showAnimationChangeMessages
        SetToggleOptionValue(_showAnimationChangeMessagesOID, ShowAnimationChangeMessages)
    ElseIf option == _wtEffectsEnabledOID
        _wtEffectsEnabled = !_wtEffectsEnabled
        SetToggleOptionValue(_wtEffectsEnabledOID, WTEffectsEnabled)
    ElseIf option == _hardcoreWTEffectsEnabledOID
        _hardcoreWTEffectsEnabled = !_hardcoreWTEffectsEnabled
        SetToggleOptionValue(_hardcoreWTEffectsEnabledOID, HardcoreWTEffectsEnabled)

    ElseIf option == _uninstallModOID
        If _uninstalling
            Return
        EndIf

        If ShowMessage("Are you sure you want to uninstall Apropos?", True)
            Database.ClearJCObjects()
            MainQuest.Stop()
            Database.Stop()
            Common.Stop()
            Widgets.Stop()
            Descriptions.Stop()
            AnimationPatchups.Stop()
            SetToggleOptionValue(_uninstallModOID, True)
            _uninstalling = True
            ForcePageReset()
            ShowMessage("Apropos quests stopped. JContainer objects released.")
        EndIf

    ElseIf option == _runAnimationPatchupsOID
        If _runningPatchups
            Return
        EndIf

        If ShowMessage("Are you sure you want to apply animation patchups?", True)
            _runningPatchups = True
            AnimationPatchups.ApplyPatchups()
        EndIf

    ElseIf option == _exportMCMOID
        If _exporting
            Return
        EndIf

        _exporting = True
        String status = ExportMCMSettings()
        _exporting = False
        ForcePageReset()
        ShowMessage(status)

    ElseIf option == _testButtonOID
        Int handle = ModEvent.Create("AproposInitWT")
        If handle
           ModEvent.PushForm(handle, PlayerRef) ;  push actor to receive W&T
           ModEvent.PushInt(handle, 9) ; level, valid values = 0 through 9
           ModEvent.PushInt(handle, 8)
           ModEvent.PushInt(handle, 7)
           ModEvent.PushBool(handle, True)
           ModEvent.PushBool(handle, True)
           ModEvent.PushBool(handle, False)
           LogToAll("Calling AproposInitWT")
           ShowMessage("Setting test W&T levels on Player...")
           ModEvent.Send(handle)
        EndIf

    ElseIf option == _importMCMOID
        If _importing
            Return
        EndIf

        If ShowMessage("Are you sure you want to import settings from " + GenerateSettingsExportPath())
            _importing = True
            String status = ImportMCMSettings()
            _importing = False
            ForcePageReset()
            ShowMessage(status)
        EndIf

    ElseIf option == _showSexDescriptionsOID
        _showSexDescriptions = !_showSexDescriptions
        SetToggleOptionValue(_showSexDescriptionsOID, ShowSexDescriptions)
    ElseIf option == _enableConsoleOutputOID
        _enableConsoleOutput = !_enableConsoleOutput
        SetToggleOptionValue(_enableConsoleOutputOID, EnableConsoleOutput)
    ElseIf option == _showVirginityLostMessagesOID
        _showVirginityLostMessages = !_showVirginityLostMessages
        SetToggleOptionValue(_showVirginityLostMessagesOID, ShowVirginityLostMessages)
    ElseIf option == _showWTChangedMessagesOID
        _showWTChangedMessages = !_showWTChangedMessages
        SetToggleOptionValue(_showWTChangedMessagesOID, ShowWTChangedMessages)
    ElseIf option == _enableSkinTexturesOID
        _enableSkinTextures = !_enableSkinTextures
        SetToggleOptionValue(_enableSkinTexturesOID, EnableSkinTextures)
    ElseIf option == _enableAfterEffectsOID
        _enableAfterEffects = !_enableAfterEffects
        SetToggleOptionValue(_enableAfterEffectsOID, EnableAfterEffects)
    ElseIf option == _enableCutScratchesOID
        _enableCutScratches = !_enableCutScratches
        SetToggleOptionValue(_enableCutScratchesOID, EnableCutScratches)
    ElseIf option == _enableDaedricScarsOID
        _enableDaedricScars = !_enableDaedricScars
        SetToggleOptionValue(_enableDaedricScarsOID, EnableDaedricScars)
    ElseIf option == _enableTearsAndSobsOID
        _enableTearsAndSobs = !_enableTearsAndSobs
        SetToggleOptionValue(_enableTearsAndSobsOID, EnableTearsAndSobs)
    ElseIf option == _widgetTestModeOID
        Widgets.TestMode = !Widgets.TestMode
        SetToggleOptionValue(_widgetTestModeOID, Widgets.TestMode)

    ElseIf option == _autoMasturbateEnabledOID
        AutoMasturbateEnabled = !AutoMasturbateEnabled
        SetToggleOptionValue(_autoMasturbateEnabledOID, AutoMasturbateEnabled)

    EndIf

EndEvent

Function ClearNPCTrackings()
    SetTextOptionValue(_clearNpcWearTearOID, "CLEARING")
    MainQuest.ClearTrackedNPCActors()
    Debug.MessageBox("NPC Wear and Tear trackings cleared!")
EndFunction

Function PatchupsExecuted()
    ShowMessage("Animation Patchups completed.")
    _runningPatchups = False
    ForcePageReset()
EndFunction

Event OnOptionDefault(Int option)
    If option == _enableDebugOID
        _debugMessagesEnabled = True
        SetToggleOptionValue(_enableDebugOID, DebugMessagesEnabled)
    ElseIf option == _enableTraceOID
        _traceMessagesEnabled = False
        SetToggleOptionValue(_enableTraceOID, TraceMessagesEnabled)        
    ElseIf option == _enableModOID
        _isModEnabled = True
        SetToggleOptionValue(_enableModOID, IsModEnabled)
    ElseIf option == _enableWearAndTearOID
        _isWearAndTearEnabled = True
        SetToggleOptionValue(_enableWearAndTearOID, IsWearAndTearEnabled)
    ElseIf option == _enableNPCWearAndTearOID
        _isNPCWearAndTearEnabled = False
        SetToggleOptionValue(_enableNPCWearAndTearOID, IsNPCWearAndTearEnabled)   
    ElseIf option == _logActorStatsOID 
        _logActorStatsEnabled = False
        SetToggleOptionValue(_logActorStatsOID, LogActorStatsEnabled) 
    ElseIf option == _showStage1MessagesOID 
        _showStage1Messages = True
        SetToggleOptionValue(_showStage1MessagesOID, ShowStage1Messages)               
    ElseIf option == _showStage2MessagesOID 
        _showStage2Messages = True
        SetToggleOptionValue(_showStage2MessagesOID, ShowStage2Messages)               
    ElseIf option == _showStage3MessagesOID 
        _showStage3Messages = True
        SetToggleOptionValue(_showStage3MessagesOID, ShowStage3Messages)               
    ElseIf option == _showStage4MessagesOID 
        _showStage4Messages = True
        SetToggleOptionValue(_showStage4MessagesOID, ShowStage4Messages)               
    ElseIf option == _showStage5MessagesOID 
        _showStage5Messages = True
        SetToggleOptionValue(_showStage5MessagesOID, ShowStage5Messages)          
    ElseIf option == _showStage6MessagesOID 
        _showStage6Messages = True
        SetToggleOptionValue(_showStage6MessagesOID, ShowStage6Messages)
    ElseIf option == _showStage7MessagesOID 
        _showStage7Messages = True
        SetToggleOptionValue(_showStage7MessagesOID, ShowStage7Messages)
    ElseIf option == _showStage8MessagesOID
        _showStage8Messages = True
        SetToggleOptionValue(_showStage8MessagesOID, ShowStage8Messages)
    ElseIf option == _showStage9MessagesOID
        _showStage9Messages = True
        SetToggleOptionValue(_showStage9MessagesOID, ShowStage9Messages)
    ElseIf option == _showStage10MessagesOID
        _showStage10Messages = True
        SetToggleOptionValue(_showStage10MessagesOID, ShowStage10Messages)
    ElseIf option == _showStage11MessagesOID
        _showStage11Messages = True
        SetToggleOptionValue(_showStage11MessagesOID, ShowStage11Messages)
    ElseIf option == _showStage12MessagesOID
        _showStage12Messages = True
        SetToggleOptionValue(_showStage12MessagesOID, ShowStage12Messages)
    ElseIf option == _showStage13MessagesOID
        _showStage13Messages = True
        SetToggleOptionValue(_showStage13MessagesOID, ShowStage13Messages)   
    ElseIf option == _showAnimationStartMessagesOID
        _showAnimationStartMessages = False
        SetToggleOptionValue(_showAnimationStartMessagesOID, ShowAnimationStartMessages)
    ElseIf option == _showAnimationChangeMessagesOID
        _showAnimationChangeMessages = False
        SetToggleOptionValue(_showAnimationChangeMessagesOID, ShowAnimationChangeMessages)        
    ElseIf option == _wtEffectsEnabledOID
        _wtEffectsEnabled = True
        SetToggleOptionValue(_wtEffectsEnabledOID, WTEffectsEnabled)
    ElseIf option == _hardcoreWTEffectsEnabledOID
        _hardcoreWTEffectsEnabled = False
        SetToggleOptionValue(_hardcoreWTEffectsEnabledOID, HardcoreWTEffectsEnabled)
    ElseIf option == _showSexDescriptionsOID
        _showSexDescriptions = True
        SetToggleOptionValue(_showSexDescriptionsOID, ShowSexDescriptions)
    ElseIf option == _enableConsoleOutputOID
        _enableConsoleOutput = True
        SetToggleOptionValue(_enableConsoleOutputOID, EnableConsoleOutput)    
    ElseIf option == _showVirginityLostMessagesOID
        _showVirginityLostMessages = True
        SetToggleOptionValue(_showVirginityLostMessagesOID, ShowVirginityLostMessages)
    ElseIf option == _showWTChangedMessagesOID
        _showWTChangedMessages = True
        SetToggleOptionValue(_showWTChangedMessagesOID, ShowWTChangedMessages)
    ElseIf option == _enableSkinTexturesOID
        _enableSkinTextures = False
        SetToggleOptionValue(_enableSkinTexturesOID, EnableSkinTextures)
    ElseIf option == _enableAfterEffectsOID
        _enableAfterEffects = True
        SetToggleOptionValue(_enableAfterEffectsOID, EnableAfterEffects)
    ElseIf option == _enableCutScratchesOID
        _enableCutScratches = True
        SetToggleOptionValue(_enableCutScratchesOID, EnableCutScratches)
    ElseIf option == _enableDaedricScarsOID
        _enableDaedricScars = True
        SetToggleOptionValue(_enableDaedricScarsOID, EnableDaedricScars)
    ElseIf option == _enableTearsAndSobsOID
        _enableTearsAndSobs = True
        SetToggleOptionValue(_enableTearsAndSobsOID, EnableTearsAndSobs)
    ElseIf option == _horizontalOffsetOID
        If Widgets.HorizontalAnchor == "Left"
            Widgets.HorizontalOffset = Widgets.DefaultHorizontalOffset
        ElseIf Widgets.HorizontalAnchor == "Center"
            Widgets.HorizontalOffset = 0
        ElseIf Widgets.HorizontalAnchor == "Right"
            Widgets.HorizontalOffset = -Widgets.DefaultHorizontalOffset
        EndIf
        SetSliderOptionValue(_horizontalOffsetOID, Widgets.HorizontalOffset, "{0} pixels")
    ElseIf option == _verticalOffsetOID
        If Widgets.VerticalAnchor == "Top"
            Widgets.VerticalOffset = Widgets.DefaultVerticalOffset
        ElseIf Widgets.VerticalAnchor == "Center"
            Widgets.VerticalOffset = 0
        ElseIf Widgets.VerticalAnchor == "Bottom"
            Widgets.VerticalOffset = -Widgets.DefaultVerticalOffset
        EndIf
        SetSliderOptionValue(_verticalOffsetOID, Widgets.VerticalOffset, "{0} pixels")
    ElseIf option == _fontSizeOID
        Widgets.FontSize = Widgets.DefaultFontSize
        SetSliderOptionValue(_fontSizeOID, Widgets.FontSize, "{0} pixels")
    ElseIf option == _playerFontSizeOID
        Widgets.PlayerFontSize = Widgets.DefaultPlayerFontSize
        SetSliderOptionValue(_playerFontSizeOID, Widgets.PlayerFontSize, "{0} pixels")
    ElseIf option == _widgetSpacingOID
        Widgets.WidgetSpacing = Widgets.DefaultWidgetSpacing
        SetSliderOptionValue(_widgetSpacingOID, Widgets.WidgetSpacing, "{0} pixels")
    ElseIf option == _horizontalAnchorOID
        Widgets.HorizontalAnchor = Widgets.DefaultHorizontalAnchor
        Widgets.HorizontalOffset = Widgets.DefaultHorizontalOffset
        SetMenuOptionValue(_horizontalAnchorOID, Widgets.HorizontalAnchor, True)
        SetSliderOptionValue(_horizontalOffsetOID, Widgets.HorizontalOffset, "{0} pixels")
    ElseIf option == _verticalAnchorOID
        Widgets.VerticalAnchor = Widgets.DefaultVerticalAnchor
        Widgets.VerticalOffset = Widgets.DefaultVerticalOffset
        SetMenuOptionValue(_verticalAnchorOID, Widgets.VerticalAnchor, True)
        SetSliderOptionValue(_verticalOffsetOID, Widgets.VerticalOffset, "{0} pixels")
    ElseIf option == _tatsColorTintOID
        SetColorOptionValue(_tatsColorTintOID, 0xFFFFFF)

    ElseIf option == _autoMasturbateEnabledOID
        AutoMasturbateEnabled = True
        SetToggleOptionValue(_autoMasturbateEnabledOID, AutoMasturbateEnabled)

    ElseIf option == _minArousalToAutoMasturbateOID
        MinArousalAutoMasturbate = 80
        SetSliderOptionValue(_minArousalToAutoMasturbateOID, MinArousalAutoMasturbate)

    EndIf
EndEvent

; Menus ....

Event OnOptionMenuOpen(Int option)
    
    String[] wearAndTearDescriptors = Database.GetAllMCMWearAndTearDescriptors()
    
    If option == _personMenuOID
        SetMenuDialogStartIndex(_personIndex)
        SetMenuDialogDefaultIndex(1)
        SetMenuDialogOptions(_personOptions)
        
    ElseIf option == _vaginalStateOID
        AproposActorAlias playerAlias = MainQuest.GetAproposActor(PlayerRef)
        If playerAlias
            Int vIndex = playerAlias.VaginalWearTearState
            SetMenuDialogStartIndex(vIndex)
            SetMenuDialogDefaultIndex(vIndex)
            SetMenuDialogOptions(wearAndTearDescriptors)
        EndIf
        
    ElseIf option == _analStateOID
        AproposActorAlias playerAlias = MainQuest.GetAproposActor(PlayerRef)
        If playerAlias 
            Int aIndex = playerAlias.AnalWearTearState
            SetMenuDialogStartIndex(aIndex)
            SetMenuDialogDefaultIndex(aIndex)
            SetMenuDialogOptions(wearAndTearDescriptors)
        EndIf
        
    ElseIf option == _oralStateOID
        AproposActorAlias playerAlias = MainQuest.GetAproposActor(PlayerRef)
        If playerAlias
            Int oIndex = playerAlias.OralWearTearState
            SetMenuDialogStartIndex(oIndex)
            SetMenuDialogDefaultIndex(oIndex)
            SetMenuDialogOptions(wearAndTearDescriptors)
        EndIf

    ElseIf option == _horizontalAnchorOID
        SetMenuDialogStartIndex(_horizontalAnchors.Find(Widgets.HorizontalAnchor))
        SetMenuDialogDefaultIndex(0)
        SetMenuDialogOptions(_horizontalAnchors)

    ElseIf option == _verticalAnchorOID
        SetMenuDialogStartIndex(_verticalAnchors.Find(Widgets.VerticalAnchor))
        SetMenuDialogDefaultIndex(0)
        SetMenuDialogOptions(_verticalAnchors)
        
    EndIf
EndEvent

Event OnOptionMenuAccept(Int option, Int index)
    
    String[] wearAndTearDescriptors = Database.GetAllMCMWearAndTearDescriptors()

    If option == _personMenuOID
        _personIndex = index
        UpdateNarrativeMode()
        SetMenuOptionValue(_personMenuOID, _personOptions[_personIndex])
        
    ElseIf option == _vaginalStateOID
        AproposActorAlias playerAlias = MainQuest.GetAproposActor(PlayerRef)
        If playerAlias
            playerAlias.OverrideVaginalWT(index, updateAbuse = True, updateCreatureAbuse = True, updateDaedricAbuse = True)

            If index == 0
                SetIntValue(PlayerRef, "hadVaginalSex", 0)
            EndIf
            SetMenuOptionValue(_vaginalStateOID, wearAndTearDescriptors[index])
            UpdateAbuseTextOptions(playerAlias)

            playerAlias.UpdateAbuseTextures()
            playerAlias.UpdateWearTearEffects()
        EndIf
        
    ElseIf option == _analStateOID
        AproposActorAlias playerAlias = MainQuest.GetAproposActor(PlayerRef)
        If playerAlias
            playerAlias.OverrideAnalWT(index, updateAbuse = True, updateCreatureAbuse = True, updateDaedricAbuse = True)
            If index == 0
                SetIntValue(PlayerRef, "hadAnalSex", 0)
            EndIf
            SetMenuOptionValue(_analStateOID, wearAndTearDescriptors[index])
            UpdateAbuseTextOptions(playerAlias)

            playerAlias.UpdateAbuseTextures()
            playerAlias.UpdateWearTearEffects()
        EndIf
        
    ElseIf option == _oralStateOID
        AproposActorAlias playerAlias = MainQuest.GetAproposActor(PlayerRef)
        If playerAlias
            playerAlias.OverrideOralWT(index, updateAbuse = True, updateCreatureAbuse = True, updateDaedricAbuse = True)
            If index == 0
                SetIntValue(PlayerRef, "hadOralSex", 0)
            EndIf        
            SetMenuOptionValue(_oralStateOID, wearAndTearDescriptors[index])
            UpdateAbuseTextOptions(playerAlias)

            playerAlias.UpdateAbuseTextures()
            playerAlias.UpdateWearTearEffects()
        EndIf

    ElseIf option == _horizontalAnchorOID
        Widgets.HorizontalAnchor = _horizontalAnchors[Index]
        Widgets.HorizontalOffset = -Widgets.DefaultHorizontalOffset * (Index - 1)
        SetMenuOptionValue(_horizontalAnchorOID, Widgets.HorizontalAnchor, True)
        SetSliderOptionValue(_horizontalOffsetOID, Widgets.HorizontalOffset, "{0} pixels")

    ElseIf option == _verticalAnchorOID
        Widgets.VerticalAnchor = _verticalAnchors[Index]
        Widgets.VerticalOffset = -Widgets.DefaultVerticalOffset * (Index - 1)
        SetMenuOptionValue(_verticalAnchorOID, Widgets.VerticalAnchor, True)
        SetSliderOptionValue(_verticalOffsetOID, Widgets.VerticalOffset, "{0} pixels")

    EndIf

EndEvent

Function UpdateAbuseTextOptions(AproposActorAlias playerAlias)
    If EnableSkinTextures
        SetTextOptionValue(_avgDaedricAbuseOID, Database.GetMCMWearAndTearDescriptor(playerAlias.AverageDaedricAbuseState))
        SetTextOptionValue(_avgGeneralAbuseOID, Database.GetMCMWearAndTearDescriptor(playerAlias.AverageAbuseState))
        SetTextOptionValue(_avgCreatureAbuseOID, Database.GetMCMWearAndTearDescriptor(playerAlias.AverageCreatureAbuseState))
    EndIf
EndFunction

; Sliders...

Event OnOptionSliderOpen(Int option)
    If option == _wtDegradeFactorOID
        SetSliderDialogStartValue(WTDegradeFactor)
        SetSliderDialogDefaultValue(1.2)
        SetSliderDialogRange(1.0, 3.0)
        SetSliderDialogInterval(0.1)
    ElseIf option == _maxWearTearAmountOID
        SetSliderDialogStartValue(MaxWearTearAmount)
        SetSliderDialogDefaultValue(2000)
        SetSliderDialogRange(500, 10000)
        SetSliderDialogInterval(200)
    ElseIf option == _frequencyWTDegradeOID
        SetSliderDialogStartValue(FrequencyWearTearDegrade)
        SetSliderDialogDefaultValue(12)
        SetSliderDialogRange(1, 24)
        SetSliderDialogInterval(1)
    ElseIf option == _fontSizeOID
        SetSliderDialogStartValue(Widgets.FontSize)
        SetSliderDialogRange(12, 20)
        SetSliderDialogInterval(1)
        SetSliderDialogDefaultValue(Widgets.DefaultFontSize)
    ElseIf option == _playerFontSizeOID
        SetSliderDialogStartValue(Widgets.PlayerFontSize)
        SetSliderDialogRange(12, 20)
        SetSliderDialogInterval(1)
        SetSliderDialogDefaultValue(Widgets.DefaultPlayerFontSize)        
    ElseIf option == _widgetSpacingOID
        SetSliderDialogStartValue(Widgets.WidgetSpacing)
        SetSliderDialogRange(-10, 10)
        SetSliderDialogInterval(1)
        SetSliderDialogDefaultValue(Widgets.DefaultWidgetSpacing)      
    ElseIf option == _horizontalOffsetOID
        SetSliderDialogStartValue(Widgets.HorizontalOffset)
        If Widgets.HorizontalAnchor == "Left"
            SetSliderDialogRange(0, 500)
            SetSliderDialogDefaultValue(Widgets.DefaultHorizontalOffset)
        ElseIf Widgets.HorizontalAnchor == "Center"
            SetSliderDialogRange(-250, 250)
            SetSliderDialogDefaultValue(0)
        ElseIf Widgets.HorizontalAnchor == "Right"
            SetSliderDialogRange(-500, 0)
            SetSliderDialogDefaultValue(-Widgets.DefaultHorizontalOffset)
        EndIf
        SetSliderDialogInterval(1)
    ElseIf option == _verticalOffsetOID
        SetSliderDialogStartValue(Widgets.VerticalOffset)
        If Widgets.VerticalAnchor == "Top"
            SetSliderDialogRange(0, 250)
            SetSliderDialogDefaultValue(Widgets.DefaultVerticalOffset)
        ElseIf Widgets.VerticalAnchor == "Center"
            SetSliderDialogRange(-125, 125)
            SetSliderDialogDefaultValue(0)
        ElseIf Widgets.VerticalAnchor == "Bottom"
            SetSliderDialogRange(-250, 0)
            SetSliderDialogDefaultValue(-Widgets.DefaultVerticalOffset)
        EndIf
        SetSliderDialogInterval(1)        
    ElseIf option == _secondsToDisplayOID
        SetSliderDialogStartValue(Widgets.SecondsToDisplay)
        SetSliderDialogRange(0.0, 10.0)
        SetSliderDialogInterval(0.5)
        SetSliderDialogDefaultValue(Widgets.DefaultSecondsToDisplay)       
    ElseIF option == _minArousalToAutoMasturbateOID
        SetSliderDialogStartValue(MinArousalAutoMasturbate)
        SetSliderDialogRange(0.0, 100.0)
        SetSliderDialogInterval(0.5)
        SetSliderDialogDefaultValue(80.0)
    EndIf
EndEvent

Event OnOptionSliderAccept(Int option, Float value)
    If option == _wtDegradeFactorOID
        _wtDegradeFactor = value
        SetSliderOptionValue(_wtDegradeFactorOID, WTDegradeFactor, "{1}")
    ElseIf option == _maxWearTearAmountOID
        _maxWearTearAmount = value As Int
        SetSliderOptionValue(_maxWearTearAmountOID, MaxWearTearAmount)
    ElseIf option == _frequencyWTDegradeOID
        _frequencyWTDegrade = value As Int
        SetSliderOptionValue(_frequencyWTDegradeOID, FrequencyWearTearDegrade)
    ElseIf option == _horizontalOffsetOID
        Widgets.HorizontalOffset = value As Int
        SetSliderOptionValue(_horizontalOffsetOID, Widgets.HorizontalOffset, "{0} pts.")
    ElseIf option == _verticalOffsetOID
        Widgets.VerticalOffset = value As Int
        SetSliderOptionValue(_verticalOffsetOID, Widgets.VerticalOffset, "{0} pts.")
    ElseIf option == _fontSizeOID
        Widgets.FontSize = value as Int
        SetSliderOptionValue(_fontSizeOID, Widgets.FontSize, "{0} pts.")
    ElseIf option == _widgetSpacingOID
        Widgets.WidgetSpacing = value as Int
        SetSliderOptionValue(_widgetSpacingOID, Widgets.WidgetSpacing, "{0} pts.")
    ElseIf option == _secondsToDisplayOID
        Widgets.SecondsToDisplay = value
        SetSliderOptionValue(_secondsToDisplayOID, Widgets.SecondsToDisplay, "{1} secs.")
    ElseIf option == _minArousalToAutoMasturbateOID
        MinArousalAutoMasturbate = value as Int
        SetSliderOptionValue(_minArousalToAutoMasturbateOID, MinArousalAutoMasturbate)
    ElseIf option == _playerFontSizeOID
        Widgets.PlayerFontSize = value as Int
        SetSliderOptionValue(_playerFontSizeOID, Widgets.PlayerFontSize, "{0} pts.")
    EndIf
EndEvent

Event OnOptionColorOpen(Int option)
    If option == _tatsColorTintOID
        SetColorDialogStartColor(TatsColorTint)
        SetColorDialogDefaultColor(0xFFFFFF)
    Endif
EndEvent
Event OnOptionColorAccept(Int option, Int color)
    If option == _tatsColorTintOID
        Log("New color accepted : " + color, True, True)
        TatsColorTint = color
        SetColorOptionValue(_tatsColorTintOID, TatsColorTint)
    Endif
EndEvent

Function NPCTrackingsCleared()
    SetTextOptionValue(_clearNpcWearTearOID, "CLEARED")
    ForcePageReset()
EndFunction

Function PCTrackingCleared()
    ForcePageReset()
EndFunction

String Function ValidateJContainers()
    If (!JContainers.isInstalled())
        String msg = "JContainers is not installed or possibly problems with SKSE or other windows (vc++) dependencies missing"
        Log(msg, True)
        Return msg
    EndIf

    Bool isValid = JContainers.APIVersion() == _requiredJCAPIVersion && JContainers.featureVersion() >= _minimumJCfeatureVersion

    If (!isValid)
        String msg = "Apropos requires version " + _requiredJCAPIVersion + " of JContainers, with a minimum feature version of " + _minimumJCfeatureVersion
        Log(msg, True)
        Return msg
    EndIf

    Return "Valid"
EndFunction


; Trace is a convenient wrapper to output 'trace' level information
; Respects the users choice to make console output silent
; To bypass Trace, just call LogToAll on AproposUtil
; Checks if Trace-level messages are enabled.
Function Trace(String mess, Bool toConsole = False, Bool toNotification = False)
    If !AreTraceMessagesEnabled()
        Return
    EndIf
    
    ToUserLog(mess)
    If toConsole && EnableConsoleOutput
        ToConsole(mess)
    EndIf
    If toNotification
        Debug.Notification(mess)
    EndIf
EndFunction

; Debug is a convenient wrapper to output 'debug' level information
; Respects the users choice to make console output silent
; To bypass Debug, just call LogToAll on AproposUtil
; Checks if Debug-level messages are enabled.
Function Debug(String mess, Bool toConsole = False, Bool toNotification = False)
    If !AreDebugMessagesEnabled()
        Return
    EndIf

    ToUserLog(mess)
    If toConsole && EnableConsoleOutput
        ToConsole(mess)
    EndIf
    If toNotification
        Debug.Notification(mess)
    EndIf
EndFunction

; Log is a convenient wrapper to output debug or trace level information
; Respects the users choice to make console output silent
; To bypass Trace, just call LogToAll on AproposUtil.
; Generally Log is used instead of Trace or Debug above
; when you have expensive log output to produce, or 
; have multiple Log statements in a group surrounded by
; If Config.AreDebugMessagesEnabled() etc.
; Does not check if Debug-level of Trace-level 
; messages are enabled or not. It is presumed that kind of checking
; is done outside of this call (AreDebugMessagesEnabled or AreTraceMessagesEnabled)
Function Log(String mess, Bool toConsole = False, Bool toNotification = False)
    ToUserLog(mess)
    If toConsole && EnableConsoleOutput
        ToConsole(mess)
    EndIf
    If toNotification
        Debug.Notification(mess)
    EndIf
EndFunction

Function LogEventInfo(String eventName, sslThreadController thread, SlaUtilScr slaUtility, Bool logActorStats = False)
    Log("---- " + eventName + " ----", EnableConsoleOutput)
    LogActors(thread, slaUtility, logActorStats)
    LogAnimation(thread)
EndFunction

Function LogActors(sslThreadController thread, SlaUtilScr slaUtility, Bool logActorStats = False)
    Actor[] actorList = thread.Positions
    Log("Actors >>", EnableConsoleOutput)
    If thread.HasCreature
        Log(thread.Creatures + " " + thread.CreatureRef.GetName(), EnableConsoleOutput)
    EndIf    
    Int i = 0
    While i < actorList.Length
        Actor anActor = actorList[i]
        String name = anActor.GetDisplayName()
        Int gender = SexLab.GetGender(anActor)
        String genderText = GetGenderName(gender)
        
        String isVictim = ""
        If thread.IsVictim(anActor)
            isVictim = "(Victim)"
        EndIf
        
        String debugText = "    [" + i + "]: " + name + ", " + genderText + " " + isVictim

        If logActorStats
            debugText += ActorStatsToString(sexLab, thread, anActor, slaUtility)
        EndIf
        Log(debugText, EnableConsoleOutput)
        i += 1
    EndWhile
EndFunction    

Function LogAnimation(sslThreadController thread)
    sslBaseAnimation animation = thread.Animation
    String tags = GetTagsAsString(animation)
    Log("Animation: '" + animation.name + "' ('" + animation.Registry +"') (" + tags + ")", EnableConsoleOutput)
EndFunction

Function LogAnimationStage(String eventName, sslThreadController thread, SlaUtilScr slaUtility, Bool logActorStats = False)
    Int stage = thread.Stage
    Log("---- " + eventName + " ---- Stage: " + stage, EnableConsoleOutput)
    If logActorStats
        LogActors(thread, slaUtility, True)
    EndIf
EndFunction

Int Function MapWearTearAmountToState(int amount)
    Return Common.MapWearTearAmountToState(amount)
EndFunction

Int Function MapWearTearStateToAmount(int wtState)
    Return Common.MapWearTearStateToAmount(wtState)
EndFunction

String Function ExportMCMSettings()
    Int exportMapId = RetainedMap()

    JMap.setInt(exportMapId, "DebugMessagesEnabled", _debugMessagesEnabled As Int)
    JMap.setInt(exportMapId, "TraceMessagesEnabled", _traceMessagesEnabled As Int)
    JMap.setInt(exportMapId, "LogActorStatsEnabled", _logActorStatsEnabled As Int)
    JMap.setInt(exportMapId, "EnableConsoleOutput", _enableConsoleOutput As Int)
    JMap.setInt(exportMapId, "IsWearAndTearEnabled", _isWearAndTearEnabled As Int)
    JMap.setInt(exportMapId, "IsNPCWearAndTearEnabled", _isNPCWearAndTearEnabled As Int)
    JMap.setFlt(exportMapId, "WTDegradeFactor", _wtDegradeFactor)
    JMap.setInt(exportMapId, "MaxWearTearAmount", _maxWearTearAmount)
    JMap.setInt(exportMapId, "FrequencyWearTearDegrade", _frequencyWTDegrade)
    JMap.setInt(exportMapId, "WTEffectsEnabled", _wtEffectsEnabled As Int)
    JMap.setInt(exportMapId, "HardcoreWTEffectsEnabled", _hardcoreWTEffectsEnabled As Int)
    JMap.setInt(exportMapId, "MinArousalAutoMasturbate", MinArousalAutoMasturbate)
    JMap.setInt(exportMapId, "AutoMasturbateEnabled", AutoMasturbateEnabled As Int)

    JMap.setInt(exportMapId, "EnableSkinTextures", _enableSkinTextures As Int)
    JMap.setInt(exportMapId, "EnableAfterEffects", _enableAfterEffects As Int)
    JMap.setInt(exportMapId, "EnableCutScratches", _enableCutScratches As Int)
    JMap.setInt(exportMapId, "EnableDaedricScars", _enableDaedricScars As Int)
    JMap.setInt(exportMapId, "EnableTearsAndSobs", _enableTearsAndSobs As Int)
    JMap.setInt(exportMapId, "TatsColorTint", TatsColorTint)

    JMap.setStr(exportMapId, "NarrativeMode", NarrativeMode)

    JMap.setInt(exportMapId, "ShowSexDescriptions", _showSexDescriptions As Int)
    JMap.setInt(exportMapId, "ShowVirginityLostMessages", _showVirginityLostMessages As Int)
    JMap.setInt(exportMapId, "ShowWTChangedMessages", _showWTChangedMessages As Int)

    JMap.setInt(exportMapId, "ShowAnimationStartMessages", _showAnimationStartMessages As Int)
    JMap.setInt(exportMapId, "ShowAnimationChangeMessages", _showAnimationChangeMessages As Int)
    JMap.setInt(exportMapId, "ShowStage1Messages", _showStage1Messages As Int)
    JMap.setInt(exportMapId, "ShowStage2Messages", _showStage2Messages As Int)
    JMap.setInt(exportMapId, "ShowStage3Messages", _showStage3Messages As Int)
    JMap.setInt(exportMapId, "ShowStage4Messages", _showStage4Messages As Int)
    JMap.setInt(exportMapId, "ShowStage5Messages", _showStage5Messages As Int)
    JMap.setInt(exportMapId, "ShowStage6Messages", _showStage6Messages As Int)
    JMap.setInt(exportMapId, "ShowStage7Messages", _showStage7Messages As Int)

    JMap.setStr(exportMapId, "Widgets.HorizontalAnchor", Widgets.HorizontalAnchor)
    JMap.setInt(exportMapId, "Widgets.HorizontalOffset", Widgets.HorizontalOffset)
    JMap.setStr(exportMapId, "Widgets.VerticalAnchor", Widgets.VerticalAnchor)
    JMap.setInt(exportMapId, "Widgets.VerticalOffset", Widgets.VerticalOffset)
    JMap.setInt(exportMapId, "Widgets.FontSize", Widgets.FontSize)
    JMap.setInt(exportMapId, "Widgets.WidgetSpacing", Widgets.WidgetSpacing)
    JMap.setInt(exportMapId, "Widgets.PlayerFontSize", Widgets.PlayerFontSize)


    String path = GenerateSettingsExportPath()
    JValue.writeToFile(exportMapId, path)
    JValue.release(exportMapId)

    ; try to immediately read it back to verify
    Int check = LoadJsonFromPath(path)
    If check
        Return "MCM Settings saved to : " + path
    Else
        Return "Could not save MCM Settings to : " + path
    EndIf
    
EndFunction


String Function ImportMCMSettings()

    String path = GenerateSettingsExportPath()

    Int importMapId = LoadJsonFromPath(path)
    If !importMapId
        Error("Previous export file not found : " + path)
        Return "Previous export file not found : " + path
    EndIf

    JValue.retain(importMapId)

    _debugMessagesEnabled = JMap.getInt(importMapId, "DebugMessagesEnabled") As Bool
    _traceMessagesEnabled = JMap.getInt(importMapId, "TraceMessagesEnabled") As Bool
    _logActorStatsEnabled = JMap.getInt(importMapId, "LogActorStatsEnabled") As Bool
    _enableConsoleOutput = JMap.getInt(importMapId, "EnableConsoleOutput") As Bool
    _isWearAndTearEnabled = JMap.getInt(importMapId, "IsWearAndTearEnabled") As Bool
    _isNPCWearAndTearEnabled = JMap.getInt(importMapId, "IsNPCWearAndTearEnabled") As Bool
    _wtDegradeFactor = JMap.getFlt(importMapId, "WTDegradeFactor")
    _maxWearTearAmount = JMap.getInt(importMapId, "MaxWearTearAmount")
    _frequencyWTDegrade = JMap.getInt(importMapId, "FrequencyWearTearDegrade")
    _wtEffectsEnabled = JMap.getInt(importMapId, "WTEffectsEnabled") As Bool
    _hardcoreWTEffectsEnabled = JMap.getInt(importMapId, "HardcoreWTEffectsEnabled") As Bool
    MinArousalAutoMasturbate = JMap.getInt(importMapId, "MinArousalAutoMasturbate")
    AutoMasturbateEnabled = JMap.getInt(importMapId, "AutoMasturbateEnabled") As Bool

    _enableSkinTextures = JMap.getInt(importMapId, "EnableSkinTextures") As Bool
    _enableAfterEffects = JMap.getInt(importMapId, "EnableAfterEffects") As Bool
    _enableCutScratches = JMap.getInt(importMapId, "EnableCutScratches") As Bool
    _enableDaedricScars = JMap.getInt(importMapId, "EnableDaedricScars") As Bool
    _enableTearsAndSobs = JMap.getInt(importMapId, "EnableTearsAndSobs") As Bool
    TatsColorTint = JMap.getInt(importMapId, "TatsColorTint")
    
    NarrativeMode = JMap.getStr(importMapId, "NarrativeMode")
    _personIndex = _personOptions.Find(NarrativeMode)

    _showSexDescriptions = JMap.getInt(importMapId, "ShowSexDescriptions") As Bool
    _showVirginityLostMessages = JMap.getInt(importMapId, "ShowVirginityLostMessages") As Bool
    _showWTChangedMessages = JMap.getInt(importMapId, "ShowWTChangedMessages") As Bool

    _showAnimationStartMessages = JMap.getInt(importMapId, "ShowAnimationStartMessages") As Bool
    _showAnimationChangeMessages = JMap.getInt(importMapId, "ShowAnimationChangeMessages") As Bool
    _showStage1Messages = JMap.getInt(importMapId, "ShowStage1Messages") As Bool
    _showStage2Messages = JMap.getInt(importMapId, "ShowStage2Messages") As Bool
    _showStage3Messages = JMap.getInt(importMapId, "ShowStage3Messages") As Bool
    _showStage4Messages = JMap.getInt(importMapId, "ShowStage4Messages") As Bool
    _showStage5Messages = JMap.getInt(importMapId, "ShowStage5Messages") As Bool
    _showStage6Messages = JMap.getInt(importMapId, "ShowStage6Messages") As Bool
    _showStage7Messages = JMap.getInt(importMapId, "ShowStage7Messages") As Bool

    Widgets.HorizontalAnchor  = JMap.getStr(importMapId, "Widgets.HorizontalAnchor")
    Widgets.HorizontalOffset = JMap.getInt(importMapId, "Widgets.HorizontalOffset")
    Widgets.VerticalAnchor = JMap.getStr(importMapId, "Widgets.VerticalAnchor")
    Widgets.VerticalOffset = JMap.getInt(importMapId, "Widgets.VerticalOffset")
    Widgets.FontSize = JMap.getInt(importMapId, "Widgets.FontSize")
    Widgets.WidgetSpacing = JMap.getInt(importMapId, "Widgets.WidgetSpacing")
    Widgets.PlayerFontSize = JMap.getInt(importMapId, "Widgets.PlayerFontSize")

    JValue.release(importMapId)

    Return "Settings imported"

EndFunction

Function PerformWidgetVisualTest()
    If Widgets.Enabled
        Widgets.PerformVisualTest()
    Else
        Debug.Notification("You need to enable Apropos Message Widgets in")
        Debug.Notification("the MCM menu in order to be able to perform tests.")
    EndIf
EndFunction

Event OnOptionKeyMapChange(Int option, Int keyCode, String conflictControl, String conflictName)
    If Option == _widgetTestHotkeyOID
        Bool accepted = True
        If conflictControl != ""
            String msg
            If conflictName != ""
                msg = "This key is already mapped to:\n\n" + conflictName + "." + conflictControl + "\n\nAre you sure you want to continue?"
            Else
                msg = "This key is already mapped to:\n\n" + conflictControl + "\n\nAre you sure you want to continue?"
            EndIf
            accepted = ShowMessage(msg, True)
        EndIf
        If accepted
            MainQuest.WidgetTestHotkey = keyCode
            SetKeymapOptionValue(_widgetTestHotkeyOID, MainQuest.WidgetTestHotkey)
        EndIf
    EndIf
EndEvent

Float Function GetReductionProbability(Int hour) 
    Float value = 0.1647826 + (-0.0047826 * (hour as Float))
    Return sslUtility.ClampFloat(value, 0.0, 1.0)
EndFunction
