ScriptName AproposDescriptionDb Extends Quest

AproposConfig Property config Auto

Import AproposUtil

Function ClearJCObjects()
    JDB.setObj(StoreKey("synonyms"), 0)
    JDB.setObj(StoreKey("arousaladjectives"), 0)
    JDB.setObj(StoreKey("weartearadjectives"), 0)
    JDB.setObj(StoreKey("actors"), 0)
    JDB.setObj(StoreKey("wearteareffects"), 0)
    JDB.setObj(StoreKey("uniqueanimations"), 0)
    JDB.setObj(StoreKey("wearteardamage"), 0)
    JDB.setObj(StoreKey("themes"), 0)
    JDB.setObj(StoreKey("themes-raw"), 0)
    ;JDB.setObj(StoreKey("weartearmap"), 0)
EndFunction

Function Initialize()
    Config.Log("Initializing Apropos Database", True)
    InitializeUniqueAnimations()
    InitializeWearTearEffects()
    InitializeWearTearDamage()
    ;InitializeWearTearMap()

    InitializeThemes()
    String maxTheme = SelectTheme(True)
    Config.Log("Selected theme for synonyms (W&T, Synonyms, Arousal) : " + maxTheme, True)

    InitializeWearAndTearAdjectives(maxTheme)
    InitializeSynonyms(maxTheme)
    InitializeArousalAdjectives(maxTheme)
    Config.Log("Apropos Database initialized", True)
EndFunction

Function InitializeThemes()
    ; load raw Themes definition into temporary JDB map
    Int rootMapId = LoadAndStoreJsonFile("Themes", "themes-raw") 

    If !rootMapId
        Return
    EndIf

    String[] allKeys = StringArrayFromJMapKeys(rootMapId)

    If !allKeys || allKeys.Length == 0
        Config.Log("Unable to read any defined themes", True, True)
        Return
    EndIf

    ; Sum all weights across all enabled Themes to determine total sample space
    Int totalWeight = 0
    Int max = 0
    String maxTheme = ""

    Int i = 0
    While i < allKeys.Length
        Int val = JDB.solveInt(QueryKey("themes-raw", allKeys[i], "weight"))

        totalWeight += val

        If val > max
            max = val
            maxTheme = allKeys[i]
        EndIf

        i +=1
    EndWhile

    Int newMapId = RetainedMap()
    JMap.setInt(newMapId, "n", totalWeight)
    JMap.setStr(newMapId, "max", maxTheme)

    i = 0
    Int k = 0
    While i < allKeys.Length
        Int val = JDB.solveInt(QueryKey("themes-raw", allKeys[i], "weight"))

        Int j = 0

        While j < val
            JMap.setStr(newMapId, (j + k) As String, allKeys[i])
            j += 1
        EndWhile

        k += j
        i +=1
    EndWhile    

    JDB.setObj(StoreKey("themes"), newMapId)

    allKeys = StringArrayFromJMapKeys(newMapId)

    i = 0
    While i < allKeys.Length
        String aKey = allKeys[i]
        If aKey == "n"
            Config.Log("themes n=" + JMap.getInt(newMapId, aKey), True)
        ElseIf aKey == "max"
            Config.Log("themes max=" + JMap.getStr(newMapId, aKey), True)
        Else
            Config.Log("theme key: " + aKey + " => " + JMap.getStr(newMapId, aKey), True)
        EndIf
        i += 1
    EndWhile

EndFunction

Function InitializeWearTearDamage()
    Int rootMapId = LoadAndStoreJsonFile("WearAndTear_Damage", "wearteardamage")

    If !rootMapId || !Config.AreDebugMessagesEnabled()
        Return
    EndIf 

    String[] allKeys = StringArrayFromJMapKeys(rootMapId)

    Int i = 0
    While i < allKeys.Length
        Float val = JDB.solveFlt(QueryKey("wearteardamage", allKeys[i]))
        Config.Log("Wear Tear Damage: " + allKeys[i] + " : " + val, True)
        i +=1
    EndWhile
EndFunction

Function InitializeUniqueAnimations()
    Int rootMapId = LoadAndStoreJsonFile("UniqueAnimations", "uniqueanimations")

    If !rootMapId || !Config.AreDebugMessagesEnabled()
        Return
    EndIf 

    String[] allKeys = StringArrayFromJMapKeys(rootMapId)

    Int i = 0
    While i < allKeys.Length
        Int val = JDB.solveInt(QueryKey("uniqueanimations", allKeys[i]))
        String disabled = ""
        If !val
            disabled = " (disabled)"
        EndIf
        Config.Log("unique animation: " + allKeys[i] + disabled, True)
        i +=1
    EndWhile

EndFunction

Function InitializeWearTearEffects()
    Int rootMapId = LoadAndStoreJsonFile("WearAndTear_Effects", "wearteareffects")

    If !rootMapId || !CheckIsJMap(rootMapId) || !Config.AreDebugMessagesEnabled()
        Return
    EndIf   

    Bool success = True

    String[] allKeys = StringArrayFromJMapKeys(rootMapId)
    Int i = 0
    While i < allKeys.Length && success
        String outerKey = allKeys[i] ; 'oral', 'oral-hc'
        
        Int mapId = JDB.solveObj(QueryKey("wearteareffects", outerKey))

        If !CheckIsJMap(mapId)
            Config.Log("Could not read values for '" + QueryKey("wearteareffects", outerKey) + "'", True)
            success = False
        Else
            String[] subKeys = StringArrayFromJMapKeys(mapId) 
            
            Int j = 0

            While j < subKeys.Length && success
                String subKey = subKeys[j] ; level4
                mapId = JDB.solveObj(QueryKey("wearteareffects", outerKey, subKey))
                If !CheckIsJMap(mapId)
                    Config.Log("Could not read values for '" + QueryKey("wearteareffects", outerKey, subKey) + "'", True)
                    success = False
                Else
                    String[] itemKeys = StringArrayFromJMapKeys(mapId)

                    Int k = 0
                    While k < itemKeys.Length
                        String itemKey = itemKeys[k]
                        Int value = JMap.getInt(mapId, itemKey)
                        Config.Log(outerKey + "." + subKey + "." + itemKey + ":" + value, True)
                        k += 1
                    EndWhile
                EndIf
                j += 1
            EndWhile
        EndIf
        i += 1
    EndWhile

EndFunction

; Function InitializeWearTearMap()
;     String topLevelMap = "weartearmap"
;     Int rootMapId = LoadAndStoreJsonFile("WearAndTear_Map", topLevelMap)

;     If !rootMapId || !CheckIsJMap(rootMapId) || !Config.AreDebugMessagesEnabled()
;         Return
;     EndIf   

;     Bool success = True

;     String[] allKeys = StringArrayFromJMapKeys(rootMapId)
;     Int i = 0
;     While i < allKeys.Length && success
;         String outerKey = allKeys[i] ; 'oral', 'oral-hc'
        
;         Int mapId = JDB.solveObj(QueryKey(topLevelMap, outerKey))

;         If !CheckIsJMap(mapId)
;             Config.Log("Could not read values for '" + QueryKey(topLevelMap, outerKey) + "'", True)
;             success = False
;         Else
;             String[] itemKeys = StringArrayFromJMapKeys(mapId)

;             Int k = 0
;             While k < itemKeys.Length
;                 String itemKey = itemKeys[k]
;                 Int value = JMap.getInt(mapId, itemKey)
;                 Config.Log(outerKey + "." + itemKey + ":" + value, True)
;                 k += 1
;             EndWhile
;         EndIf
;         i += 1
;     EndWhile
; EndFunction

Function InitializeSynonyms(string theme)
    Int rootMapId = LoadAndStoreJsonFile("Synonyms", "synonyms", theme)

    If !rootMapId || !Config.AreDebugMessagesEnabled()
        Return
    EndIf

    String[] allKeys = StringArrayFromJMapKeys(rootMapId)
    
    Int index = 0
    
    Bool success = True
    
    While ((index < allKeys.Length) && success)
        String currentTokenName = allKeys[index]
        Int arrayId = JDB.solveObj(QueryKey("synonyms", currentTokenName))
        If !CheckIsJArray(arrayId)
            Config.Log("Could not read values for " + currentTokenName, True)
            success = False
        Else
            String[] values = StringArrayFromJArray(arrayId)
            Config.Log("Values for " + currentTokenName + " : " + StringArrayToString(values), True)
        EndIf
        index += 1
    EndWhile
    
    If success
        Config.Log("Synonyms successfully read, parsed and stored.", True)
    Else
        Config.Log("Could not read all synonyms.", True)
    EndIf
EndFunction

Function InitializeArousalAdjectives(string theme)
    Int rootMapId = LoadAndStoreJsonFile("Arousal_Descriptors", "arousaladjectives", theme)

    If !rootMapId || !Config.AreDebugMessagesEnabled()
        Return
    EndIf

    Bool success = True 

    String[] allKeys = StringArrayFromJMapKeys(rootMapId)

    Int i = 0
    While (i < allKeys.Length && success)
        String outerKey = allKeys[i]
        Int mapId = JDB.solveObj(QueryKey("arousaladjectives", outerKey))

        Config.Log("Values for " + outerKey, True)

        If !CheckIsJMap(mapId)
            Config.Log("Could not read values for '" + QueryKey("arousaladjectives", outerKey) + "'", True)
            success = False
        Else

            String[] subKeys = StringArrayFromJMapKeys(mapId)
            
            Int j = 0

            While (j < subKeys.Length && success)
                String subKey = subKeys[j]
                Int arrayId = JDB.solveObj(QueryKey("arousaladjectives", outerKey, subKey))
                If !CheckIsJArray(arrayId)
                    Config.Log("Could not read values for '" + QueryKey("arousaladjectives", outerKey, subKey) + "'", True)
                    success = False
                EndIf
                String[] values = StringArrayFromJArray(arrayId)
                Config.Log("Values for " + subKey + " : " + StringArrayToString(values), True)
                j += 1
            EndWhile

        EndIf

        i += 1
    EndWhile

    If success
        Config.Log("Arousal Adjectives successfully read, parsed and stored.", True)
    Else
        Config.Log("Could not read all Arousal Adjectives.", True)
    EndIf    

EndFunction

Function InitializeWearAndTearAdjectives(string theme)
    Int rootMapId = LoadAndStoreJsonFile("WearAndTear_Descriptors", "weartearadjectives", theme)

    If !rootMapId || !Config.AreDebugMessagesEnabled()
        Return
    EndIf

    Bool success = True 

    Int wearTearMapId = JDB.solveObj(QueryKey("weartearadjectives.descriptors"))
    
    If !CheckIsJMap(wearTearMapId)
        Config.Log("Could not read values for '" + QueryKey("weartearadjectives.descriptors") + "'", True)
        success = False
    Else

        String[] allKeys = StringArrayFromJMapKeys(wearTearMapId)
        
        Int index = 0

        While (index < allKeys.Length && success)
            String currentKey = allKeys[index]
            Int arrayId = JDB.solveObj(QueryKey("weartearadjectives.descriptors", currentKey))
            If !CheckIsJArray(arrayId)
                Config.Log("Could not read values for '" + QueryKey("weartearadjectives.descriptors", currentKey) + "'", True)
                success = False
            EndIf
            String[] values = StringArrayFromJArray(arrayId)
            Config.Log("Values for " + currentKey + " : " + StringArrayToString(values), True)
            index += 1
        EndWhile

    EndIf

    If success
        Config.Log("Synonyms successfully read, parsed and stored.", True)
    Else
        Config.Log("Could not read all synonyms.", True)
    EndIf    

EndFunction

Int Function GetWearTearDamage(String damageClass, Bool isRapeOrAggressive)
    If Config.AreTraceMessagesEnabled()
        Config.Log("AproposDescriptionDb.GetWearTearDamage(damageClass="+damageClass+",isRapeOrAggressive="+isRapeOrAggressive+")")
    EndIf

    Float amt = JDB.solveFlt(QueryKey("wearteardamage", damageClass), -1.0)
    If amt == -1.0
        Config.Log("WARNING: AproposDescriptionDb.GetWearTearDamage() - " + damageClass + " unrecognized. Using a value of 10.0 for damage", True)
        amt = 10.0
    EndIf

    If isRapeOrAggressive
        Float factor = JDB.solveFlt(QueryKey("wearteardamage", "AggressiveOrRape"), 1.0)
        Return (amt * factor) As Int
    EndIf
    Return amt As Int
EndFunction

String Function GetRandomWearAndTearDescriptor(Int wearTearLevel)
    String lvl = "level" + SslUtility.ClampInt(wearTearLevel, 0, 9)
    Int arrayId = JDB.solveObj(QueryKey("weartearadjectives.descriptors", lvl))
    If !CheckIsJArray(arrayId)
        Return DummyArray()
    EndIf
    String[] values = StringArrayFromJArray(arrayId)
    Int random = Utility.RandomInt(0, values.Length - 1)
    Return values[random]
EndFunction

String Function GetMCMWearAndTearDescriptor(Int wearTearLevel)
    wearTearLevel = SslUtility.ClampInt(wearTearLevel, 0, 9)
    Int arrayId = JDB.solveObj(QueryKey("weartearadjectives.descriptors-mcm"))
    If !CheckIsJArray(arrayId)
        Return DummyArray()
    EndIf
    String[] values = StringArrayFromJArray(arrayId)
    Return values[wearTearLevel]
EndFunction

String[] Function GetAllMCMWearAndTearDescriptors()
    Int arrayId = JDB.solveObj(QueryKey("weartearadjectives.descriptors-mcm"))
        If !CheckIsJArray(arrayId)
        Return DummyArray()
    EndIf
    Return StringArrayFromJArray(arrayId)
EndFunction

String Function RandomSynonym(String tokenName)
    Int arrayId = JDB.solveObj(QueryKey("synonyms", tokenName))
    If arrayId == 0
        Config.Log("Could not find synonym array for " + tokenName, True)
        Return ""
    EndIf
    
    Int random = Utility.RandomInt(0, JArray.count(arrayId) - 1)
    Return JArray.getStr(arrayId, random)
EndFunction

String Function RandomArousal(String tokenName, Int arousalLevel)
    String lvl = "level" + SslUtility.ClampInt(arousalLevel, 0, 4)
    Int arrayId = JDB.solveObj(QueryKey("arousaladjectives", tokenName, lvl))
    If arrayId == 0
        Config.Log("Could not find arousal adjective array for " + tokenName, True)
        Return ""
    EndIf
    
    Int random = Utility.RandomInt(0, JArray.count(arrayId) - 1)
    Return JArray.getStr(arrayId, random)
EndFunction

String[] Function AllSynonymTokenNames()
    Return AllMapKeysAsStringArray("synonyms")
EndFunction

String[] Function AllMapKeysAsStringArray(String storageKey)
    Int mapId = JDB.solveObj(QueryKey(storageKey))
    If !CheckIsJMap(mapId)
        Config.Log("Could not read storageKey or storageKey is not a map : " + QueryKey(storageKey), True)
        Return DummyArray()
    EndIf
    Return StringArrayFromJMapKeys(mapId)
EndFunction

Bool Function IsUnique(String animation)
    Bool isUnique = JDB.solveInt(QueryKey("uniqueanimations", animation))
    If isUnique && Config.AreDebugMessagesEnabled()
        Config.Log("animation " + animation + " is listed as unique", True)
    EndIf
    Return isUnique
EndFunction

String Function SelectTheme(Bool selectMaxTheme = False)
    If selectMaxTheme
        String maxTheme = JDB.solveStr(QueryKey("themes", "max"))
        Config.Log("Selected (max) theme: " + maxTheme)
        Return maxTheme
    Else
        Int n = JDB.solveInt(QueryKey("themes", "n"))
        If n
            Int i = Utility.RandomInt(0, n - 1)
            String theme = JDB.solveStr(QueryKey("themes", (i As String)))
            Config.Log("Selected theme: " + theme)
            Return theme
        Else
            Config.Trace("Could not find any configured themes")
            Return ""
        EndIf        
    EndIf

EndFunction

; sexPart = Oral, Anal, Vaginal, Boobjob, or GangBang
String Function RetrieveDescription(SslThreadController thread, String subjectPrefix, String partner, Bool isActorVictim, Bool isOrgasm, String sexPart, String narrativeVoice)
    String animation = thread.Animation.Registry

    If config.AreTraceMessagesEnabled()
        Config.Log("AproposDescriptionDb.RetrieveDescription(animation=" + animation + ",subjectPrefix=" + subjectPrefix + ",partner=" + partner + ",isActorVictim=" + isActorVictim + ",isOrgasm=" + isOrgasm + ",sexPart=" + sexPart+",narrativeVoice="+narrativeVoice+")")
    EndIf

    String rape = ""
    If (isActorVictim)
        rape = "_Rape"
    EndIf
    
    String orgasm = ""
    If (isOrgasm)
        orgasm = "_Orgasm"
    EndIf

    Bool isUnique = IsUnique(animation)
    
    sexPart = "_" + sexPart
    partner = "_" + partner
    animation = "_" + animation

    String fullPath
    Bool foundUniqueAnimationFile = False

    String theme = SelectTheme()

    If isUnique
        fullPath = GenerateFullPath(subjectPrefix + animation + rape + orgasm, subjectPrefix + animation, theme)
        If theme && !CheckFileExists(fullPath)
            ; fallback
            fullPath = GenerateFullPath(subjectPrefix + animation + rape + orgasm, subjectPrefix + animation)
        EndIf
        foundUniqueAnimationFile = CheckFileExists(fullPath)
    EndIf
    
    If !foundUniqueAnimationFile
        fullPath = GenerateFullPath(subjectPrefix + partner + sexPart + rape + orgasm, subjectPrefix + partner, theme)
        If theme && !CheckFileExists(fullPath)
            ; fallback
            fullPath = GenerateFullPath(subjectPrefix + partner + sexPart + rape + orgasm, subjectPrefix + partner)
        EndIf
    EndIf

    ToUserLog("Using file '" + fullPath + "'")
    
    Int arrayId = ReadDescriptionFileWithNarrativeVoice(fullPath, narrativeVoice)    
    If (arrayId == 0)
        Return ""
    EndIf

    Return GetStringFromArray(arrayId, fullPath, narrativeVoice)
EndFunction

String Function RetrieveFemaleWearTearChangedDescription(String change, String sexPart, String narrativeVoice)
    If config.AreTraceMessagesEnabled()
        Config.Log("AproposDescriptionDb.RetrieveFemaleWearTearChangedDescription(sexPart=" + sexPart+",narrativeVoice="+narrativeVoice+")")
    EndIf

    String theme = SelectTheme()

    sexPart = "_" + sexPart
    String fullPath = GenerateFullPath("FemaleActor_WearTear" + change + sexPart, "FemaleActor", theme)
    If theme && !CheckFileExists(fullPath)
        ; fallback
        fullPath = GenerateFullPath("FemaleActor_WearTear" + change + sexPart, "FemaleActor")
    EndIf
    
    Int arrayId = ReadDescriptionFileWithNarrativeVoice(fullPath, narrativeVoice)    
    If (arrayId == 0)
        Return ""
    EndIf

    Return GetStringFromArray(arrayId, fullPath, narrativeVoice)
EndFunction

String Function RetrieveStageProgressionDescription(SslThreadController thread, String subjectPrefix, Bool isActorVictim, String partner, String sexPart, String narrativeVoice, Int stage)
    String animation = thread.Animation.Registry
    If config.AreTraceMessagesEnabled()
        Config.Log("AproposDescriptionDb.RetrieveStageProgressionDescription(animation=" + animation + ",subjectPrefix=" + subjectPrefix + ",isActorVictim=" + isActorVictim + ",partner=" + partner + ",sexPart=" + sexPart+",narrativeVoice="+narrativeVoice+ ",stage=" + stage + ")")
    EndIf

    Bool isUnique = IsUnique(animation)

    sexPart = "_" + sexPart
    partner = "_" + partner
    animation = "_" + animation

    String rape = ""
    If isActorVictim
        rape = "_Rape"
    EndIf    

    String theme = SelectTheme()

    String stageStr = "_Stage" + stage

    String fullPath
    Bool foundUniqueAnimationFile = False

    If isUnique
        fullPath = GenerateFullPath(subjectPrefix + animation + rape + stageStr, subjectPrefix + animation, theme)
        If theme && !CheckFileExists(fullPath)
            ; fallback
            fullPath = GenerateFullPath(subjectPrefix + animation + rape + stageStr, subjectPrefix + animation)
        EndIf
        foundUniqueAnimationFile = CheckFileExists(fullPath)
    EndIf
    
    If !foundUniqueAnimationFile
        fullPath = GenerateFullPath(subjectPrefix + partner + sexPart + rape + stageStr, subjectPrefix + partner, theme)
        If theme && !CheckFileExists(fullPath)
            ; fallback
            fullPath = GenerateFullPath(subjectPrefix + partner + sexPart + rape + stageStr, subjectPrefix + partner)
        EndIf
    EndIf

    ToUserLog("Using file '" + fullPath + "'")
    
    Int arrayId = ReadDescriptionFileWithNarrativeVoice(fullPath, narrativeVoice)    
    If !arrayId
        Return ""
    EndIf

    Return GetStringFromArray(arrayId, fullPath, narrativeVoice)
EndFunction

String Function RetrieveVirginityLostDescription(String subjectPrefix, String sexPart, String narrativeVoice, Bool isActorVictim)
    If config.AreTraceMessagesEnabled()
        Config.Log("AproposDescriptionDb.RetrieveVirginityLostDescription(subjectPrefix=" + subjectPrefix + ",isActorVictim=" + isActorVictim +",sexPart=" + sexPart+",narrativeVoice="+narrativeVoice+")")
    EndIf
    
    String rape = ""
    If isActorVictim
        rape = "_Rape"
    EndIf

    String theme = SelectTheme()
    
    sexPart = "_" + sexPart
    String fullPath = GenerateFullPath(subjectPrefix + "_VirginityLost" + sexPart + rape, subjectPrefix, theme)
    If theme && !CheckFileExists(fullPath)
        ; fallback
        fullPath = GenerateFullPath(subjectPrefix + "_VirginityLost" + sexPart + rape, subjectPrefix)
    EndIf
    
    Int arrayId = ReadDescriptionFileWithNarrativeVoice(fullPath, narrativeVoice)    
    If !arrayId
        Return ""
    EndIf
    
    Return GetStringFromArray(arrayId, fullPath, narrativeVoice)
EndFunction

String Function RetrieveMasturbationDescription(SslThreadController thread, String subjectPrefix, Bool isOrgasm, String narrativeVoice, Int stageNumber = 0)
    If config.AreTraceMessagesEnabled()
        Config.Log("AproposDescriptionDb.RetrieveMasturbationDescription(animation=" + thread.Animation.Registry + ",subjectPrefix=" + subjectPrefix +",narrativeVoice="+narrativeVoice+")")
    EndIf

    String orgasm = ""
    If isOrgasm
        orgasm = "_Orgasm"
    EndIf

    String stage = ""
    If stageNumber > 0
        stage = "_Stage" + stageNumber
    EndIf

    String theme = SelectTheme()
    
    String fullPath = GenerateFullPath(subjectPrefix + "_Masturbation" + orgasm + stage, subjectPrefix, theme)
    If theme && !CheckFileExists(fullPath)
        ; fallback
        fullPath = GenerateFullPath(subjectPrefix + "_Masturbation" + orgasm + stage, subjectPrefix)
    EndIf
    
    Int arrayId = ReadDescriptionFileWithNarrativeVoice(fullPath, narrativeVoice)    
    If !arrayId
        Return ""
    EndIf
    
    Return GetStringFromArray(arrayId, fullPath, narrativeVoice)
EndFunction

Int Function ReadDescriptionFileWithNarrativeVoice(String fullPath, String narrativeVoice)
    Config.Log("ReadDescriptionFileWithNarrativeVoice: fullPath=" + fullPath + ", narrativeVoice=" + narrativeVoice)
    If !CheckFileExists(fullPath)
        Return 0
    EndIf
    
    Int rootMap = JValue.readFromFile(fullPath)
    
    If !rootMap
        Config.Log("Problems parsing file '" + fullPath + "'. Check Json validity at jsonlint.com", True)
        Return 0
    EndIf   
    
    If !JMap.hasKey(rootMap, narrativeVoice)
        Config.Log("Could not find narrative voice '" + narrativeVoice + "' for file '" + fullPath + "'. Also check Json validity at jsonlint.com", True)
        Return 0
    EndIf
    
    Int arrayId = JMap.getObj(rootMap, narrativeVoice)

    ; Config.Log("ReadDescriptionFileWithNarrativeVoice: arrayId="+ arrayId)
    Return arrayId
EndFunction

String Function GetStringFromArray(Int arrayId, String fullPath, String narrativeVoice)
    String result = ""
    Int count = JArray.count(arrayId)
    If count == 0
        If config.AreDebugMessagesEnabled()
            Config.Log("Apparently there are no descriptions in the file '" + fullPath + "' for the narrative voice " + narrativeVoice, True)
        EndIf        
    Else
        Int random = Utility.RandomInt(0, count - 1)
        result = JArray.getStr(arrayId, random)
        If !result
            If config.AreDebugMessagesEnabled()
                Config.Log("Apparently there are no descriptions in the file '" + fullPath + "' for the narrative voice " + narrativeVoice, True)
            EndIf
        ElseIf config.AreTraceMessagesEnabled()
            Config.Log("AproposDescriptionDb.GetStringFromArray() returning: " + result)
        EndIf        
    EndIf
    Return result
EndFunction

String Function RetrieveFemaleMasturbationDescription(SslThreadController thread, Bool isOrgasm, String narrativeVoice, Int stageNumber = 0)
    Return RetrieveMasturbationDescription(thread, "FemaleActor", isOrgasm, narrativeVoice, stageNumber)
EndFunction

String Function RetrieveMaleMasturbationDescription(SslThreadController thread, Bool isOrgasm, String narrativeVoice, Int stageNumber = 0)
    Return RetrieveMasturbationDescription(thread, "MaleActor", isOrgasm, narrativeVoice, stageNumber)
EndFunction

String Function RetrieveFemaleStageProgressionDescription(SslThreadController thread, String partner, Bool isVictim, String sexPart, String narrativeVoice, Int stage)
    Return RetrieveStageProgressionDescription(thread, "FemaleActor", isVictim, partner, sexPart, narrativeVoice, stage)
EndFunction

String Function RetrieveMaleStageProgressionDescription(SslThreadController thread, String partner, Bool isVictim, String sexPart, String narrativeVoice, Int stage)
    Return RetrieveStageProgressionDescription(thread, "MaleActor", isVictim, partner, sexPart, narrativeVoice, stage)
EndFunction

String Function RetrieveFemaleVirginityLostDescription(Bool isActorVictim, String sexPart, String narrativeVoice)
    Return RetrieveVirginityLostDescription("FemaleActor", sexPart, narrativeVoice, isActorVictim)   
EndFunction

String Function RetrieveMaleVirginityLostDescription(Bool isActorVictim, String sexPart, String narrativeVoice)
    Return RetrieveVirginityLostDescription("MaleActor", sexPart, narrativeVoice, isActorVictim)   
EndFunction

; sexPart = Oral, Anal, Vaginal, Boobjob, or GangBang
String Function RetrieveFemaleDescription(SslThreadController thread, String partner, Bool isActorVictim, Bool isOrgasm, String sexPart, String narrativeVoice)
    Return RetrieveDescription(thread, "FemaleActor", partner, isActorVictim, isOrgasm, sexPart, narrativeVoice)   
EndFunction

; sexPart = Oral, Anal, Vaginal, Boobjob, or GangBang
String Function RetrieveMaleDescription(SslThreadController thread, String partner, Bool isActorVictim, Bool isOrgasm, String sexPart, String narrativeVoice)
    Return RetrieveDescription(thread, "MaleActor", partner, isActorVictim, isOrgasm, sexPart, narrativeVoice)
EndFunction

String Function ReplaceTokens(String source, Int tokenMapId)
    ; If config.AreTraceMessagesEnabled()
    ;     Config.Log("TokenNames : " + StringArrayToString(StringArrayFromJArray(JMap.allKeys(tokenMapId))))
    ;     Config.Log("TokenValues : " + StringArrayToString(StringArrayFromJArray(JMap.allValues(tokenMapId))))
    ; EndIf
    
    String result = ""
    
    While StringUtil.GetLength(source) > 0
        ;ToUserLog("source = " + source)
        Int sourceLength = StringUtil.GetLength(source)
        
        Int indexTokenStart = StringUtil.Find(source, "{")
        Int indexTokenEnd = StringUtil.Find(source, "}")
        
        ;ToUserLog("0. indexTokenStart = " + indexTokenStart + ", indexTokenEnd = " + indexTokenEnd)
        
        If indexTokenStart != -1 && indexTokenEnd != -1
            String tokenName = Substring(source, indexTokenStart, indexTokenEnd - indexTokenStart + 1);
            ;ToUserLog("1. tokenName = " + tokenName)
            ;ToUserLog("2. tokenIndex = " + tokenIndex)
            String beforeToken = Substring(source, 0, indexTokenStart);
            String afterToken = Substring(source, indexTokenEnd + 1, sourceLength - (indexTokenEnd + 1));
            ;ToUserLog("3. beforeToken = " + beforeToken)
            ;ToUserLog("4. afterToken = " + afterToken)
            If (JMap.hasKey(tokenMapId, tokenName))
                String tokenValue = JMap.getStr(tokenMapId, tokenName)
                ;ToUserLog("Replacing '" + tokenName + "' with value '" + tokenValue + "'")
                If (tokenValue == "")
                    result = result + TrimEnd(beforeToken)
                Else
                    result = result + beforeToken + tokenValue;
                EndIf
                ;ToUserLog("5. result = " + result)
            Else
                result = result + beforeToken
                ;ToUserLog("6. result = " + result)					
            EndIf                
            
            source = afterToken;
        Else
            result = result + source;
            ;ToUserLog("7. result = " + result)
            source = "";
        EndIf
        
    EndWhile

    If Config.AreTraceMessagesEnabled()
        Config.Log("AproposDescriptionDb.ReplaceTokens returned : " + result)
    EndIf
    
    Return result
    
EndFunction
