Scriptname AproposWidgetManager extends Quest

AproposConfig Property Config Auto

Import AproposUtil

AproposMessageWidget Property MessageWidget0 Auto
AproposMessageWidget Property MessageWidget1 Auto
AproposMessageWidget Property MessageWidget2 Auto
AproposMessageWidget Property MessageWidget3 Auto
AproposMessageWidget Property MessageWidget4 Auto
AproposMessageWidget Property MessageWidget5 Auto
AproposMessageWidget Property MessageWidget6 Auto
AproposMessageWidget Property MessageWidget7 Auto
AproposMessageWidget Property MessageWidget8 Auto
AproposMessageWidget Property MessageWidget9 Auto
AproposMessageWidget[] Property Widgets Auto Hidden

; Inbound queue request data
String[] _msgSections
String[] _msgTexts
; Once we reach the end of the arrays, we start over.
Int _firstMsg
Int _msgCount
Bool _isRunning

String _horizontalAnchor
String _verticalAnchor
Int _widgetSpacing
Int _horizontalOffset
Int _verticalOffset
Bool _testMode
Int _fontSize
Int _playerFontSize
Float _alpha
Bool _initialized
Bool _isEnabled
Bool _isVisible
Float _secondsToDisplay

Event OnInit()
	Config.Log("AproposWidgetManager.OnInit", True)
    SetUpWidgets()
    InitWidgets()
    _msgTexts         = new String[128]
    _msgSections      = new String[128]
EndEvent

; This function never release the lock on "self"
Function EnsurePumpIsRunning()
	If !_isRunning
		_isRunning = True
		RegisterForSingleUpdate(0.1) ; <-- this function has the same self as this script, so calling RegisterForSingleUpdate()
	EndIf                            ; will not unlock the script. (However, once the EnsurePumpIsRunning() function has finished,
EndFunction                          ; other threads CAN sneak in during the wait time before the update).

Int Property MaxQueueLength = 128 AutoReadOnly

; This function never release the lock on "self"
; (mess) -> queues messages for W&T
Function EnqueueDescriptionMessage(String mess, String section)
	Int index = (_firstMsg + _msgCount) % MaxQueueLength
	_msgCount += 1
	_msgSections[index] = section
	_msgTexts[index] = mess
	EnsurePumpIsRunning()
EndFunction

; Displays message to player widget directly, without queuing, threading logic.
Function DisplayPlayerDescriptionMessage(String mess)
    Widgets[0].DisplayPlayerMessage(mess)
    UpdatePosition(0)
EndFunction

Event OnUpdate()
	While _msgCount != 0
		int index = _firstMsg
		_firstMsg += 1
		If _firstMsg > (MaxQueueLength - 1)
			_firstMsg = 0
		EndIf
		_msgCount -= 1

        ; If Config.AreTraceMessagesEnabled()
        ;     Config.Log("OnUpdate... _firstMsg : " + _firstMsg + ", _msgCount : " + _msgCount, True)
        ; EndIf   

		String section = _msgSections[index]
		String text = _msgTexts[index]

        ; If Config.AreDebugMessagesEnabled()
        ;     Config.Log("OnUpdate... section : " + section + ", text " + text, True)
        ; EndIf        

		AproposMessageWidget widget = FindWidget(section)

		If widget
			widget.DisplayMessage(text)
            UpdatePosition(widget.WidgetIndex)
		Else
			Error("Could not find an available widget for message: '" + text + "'")
		EndIf

	EndWhile
	_isRunning = False
EndEvent

; used to test graphical functionality of widget only. Unrelated to queuing system.
Function PerformVisualTest()
    If !_initialized
        Return
    EndIf
    
    If !TestMode
        Debug.Notification("Unable to run Apropos widget test because Test mode is not enabled in MCM.")
    Else
        Config.Log("AproposWidgetManager.PerformVisualTest", True)

        UpdateAllWidgetPositions()
        Debug.Notification(CreateTestSentence(1, 5))

        Int i = 0
        While i < Widgets.Length
            Widgets[i].Test(" : " + CreateTestSentence(10, 15) + ".")
            Debug.Notification(CreateTestSentence(1, 5))
            i += 1
        EndWhile
    EndIf
EndFunction

AproposMessageWidget Function FindWidget(String section)
	Int i = 1 
	While i < Widgets.Length
		AproposMessageWidget widget = Widgets[i]
        ;Config.Log("FindWidget() " + widget.DebugText)
        If widget.IsEnabled && widget.WidgetSection == section
            Return widget
        EndIf
		i += 1
	EndWhile
	Return None
EndFunction

Function SetUpWidgets()
	Config.Debug("AproposWidgetManager.SetUpWidgets", True)
	Widgets = New AproposMessageWidget[7]
	Widgets[0] = MessageWidget0
	Widgets[1] = MessageWidget1
	Widgets[2] = MessageWidget2
	Widgets[3] = MessageWidget3
	Widgets[4] = MessageWidget4
	Widgets[5] = MessageWidget5
	Widgets[6] = MessageWidget6

    Widgets[0].IsEnabled = True
    Widgets[1].IsEnabled = True
    Widgets[2].IsEnabled = True
    Widgets[3].IsEnabled = True
    Widgets[4].IsEnabled = True
    Widgets[5].IsEnabled = True
    Widgets[6].IsEnabled = True

    Widgets[0].WidgetSection = "Player"
    Widgets[1].WidgetSection = "NPC"
    Widgets[2].WidgetSection = "NPC"
    Widgets[3].WidgetSection = "NPC"
    Widgets[4].WidgetSection = "MISC"
    Widgets[5].WidgetSection = "MISC"
    Widgets[6].WidgetSection = "MISC"

    Int i = 0 
    While i < Widgets.Length
        AproposMessageWidget widget = Widgets[i]
        widget.WidgetIndex = i
        i += 1
    EndWhile

EndFunction

Function GameLoaded()
	Config.Debug("AproposWidgetManager.GameLoaded", True)

	If !Widgets
		SetUpWidgets()
	EndIf

EndFunction

Function InitWidgets()
    Config.Log("AproposWidgetManager.InitWidgets", True)

    Self.Enabled = True
    Self.Visible = False
    _testMode = False
    _fontSize = DefaultFontSize
    _playerFontSize = DefaultPlayerFontSize
    _horizontalAnchor = DefaultHorizontalAnchor
    _verticalAnchor = DefaultVerticalAnchor
    _horizontalOffset = DefaultHorizontalOffset
    _verticalOffset = DefaultVerticalOffset
    _widgetSpacing = DefaultWidgetSpacing
    _secondsToDisplay = DefaultSecondsToDisplay
    _alpha = 100.0
    UpdateWidgets()
    _initialized = True
EndFunction

Function UpdateWidgets()
	UpdateStyle()
	UpdateAllWidgetPositions()
EndFunction

Function UpdateStyle()
    Self.Alpha = _alpha
    Self.FontSize = _fontSize
    Self.PlayerFontSize = _playerFontSize
    ;Self.Visible = _isVisible
EndFunction

Function Stop()
	Self.Enabled = FALSE
	Int i = 0
	While i < Widgets.Length
		Widgets[i].Stop()
		i +=1 
	EndWhile
	Parent.Stop()
EndFunction

Bool Property TestMode
    Bool Function get()
        Return _testMode
    EndFunction

    Function set(Bool a_val)
        _testMode = a_val
    EndFunction
EndProperty

Bool Property Enabled
	Bool Function get()
		Return _isEnabled
	EndFunction

	Function set(Bool a_val)
        If Config.AreTraceMessagesEnabled()
            Config.Log("AproposWidgetManager.Enabled.set(" + a_val + ")", True)
        EndIf
		If a_val
			_isEnabled = a_val
			Self.Visible = _isEnabled
		Else
			Self.Visible = a_val
			_isEnabled = a_val
		EndIf                        
	EndFunction
EndProperty

Bool Property Visible
	Bool Function Get()
		Return _isVisible
	EndFunction

	Function Set(Bool a_val)
        ; If Config.AreTraceMessagesEnabled()
        ;     Config.Log("AproposWidgetManager.Visible.set(" + a_val + ")", True)
        ; EndIf
		If _isEnabled
			 _isVisible = a_val

			 Int i = 0
			 While i < Widgets.Length
			 	Widgets[i].Visible = a_val
			 	i += 1
             EndWhile
        EndIf
    EndFunction
EndProperty

Float Property SecondsToDisplay
    Float Function get()
        Return _secondsToDisplay
    EndFunction

    Function set(Float a_val)
        ; If Config.AreTraceMessagesEnabled()
        ;     Config.Log("AproposWidgetManager.SecondsToDisplay.set(" + a_val + ")", True)
        ; EndIf
        _secondsToDisplay = a_val
        Int i = 0
        While i < Widgets.Length
            Widgets[i].SecondsToDisplay = a_val
            i += 1
        EndWhile
        PerformVisualTest()        
    EndFunction
EndProperty

Int Property FontSize
    Int Function get()
        Return _fontSize
    EndFunction

    Function set(Int a_val)
        ; If Config.AreTraceMessagesEnabled()
        ;     Config.Log("AproposWidgetManager.FontSize.set(" + a_val + ")", True)
        ; EndIf
        _fontSize = a_val
		Int i = 1
		While i < Widgets.Length
			Widgets[i].TextSize = a_val
			i += 1
		EndWhile
        UpdateAllWidgetPositions()
        PerformVisualTest()
    EndFunction
EndProperty

Int Property PlayerFontSize
    Int Function get()
        Return _playerFontSize
    EndFunction

    Function set(Int a_val)
        ; If Config.AreTraceMessagesEnabled()
        ;     Config.Log("AproposWidgetManager.PlayerFontSize.set(" + a_val + ")", True)
        ; EndIf
        _playerFontSize = a_val
        Widgets[0].TextSize = a_val
        UpdateAllWidgetPositions()
        PerformVisualTest()
    EndFunction
EndProperty

Int Property WidgetSpacing
    Int Function get()
        Return _widgetSpacing
    EndFunction

    Function set(Int a_val)
        ; If Config.AreTraceMessagesEnabled()
        ;     Config.Log("AproposWidgetManager.WidgetSpacing.set(" + a_val + ")", True)
        ; EndIf
        _widgetSpacing = a_val
        UpdateAllWidgetPositions()
        PerformVisualTest()
    EndFunction
EndProperty

Int Property HorizontalOffset
    Int Function get()
        Return _horizontalOffset
    EndFunction

    Function set(Int a_val)
        ; If Config.AreTraceMessagesEnabled()
        ;     Config.Log("AproposWidgetManager.HorizontalOffset.set(" + a_val + ")", True)
        ; EndIf
        _horizontalOffset = a_val
        UpdateAllWidgetPositions()
        PerformVisualTest()
    EndFunction
EndProperty

Int Property VerticalOffset
    Int Function get()
        Return _verticalOffset
    EndFunction

    Function set(Int a_val)
        ; If Config.AreTraceMessagesEnabled()
        ;     Config.Log("AproposWidgetManager.VerticalOffset.set(" + a_val + ")", True)
        ; EndIf
        _verticalOffset = a_val
        UpdateAllWidgetPositions()
        PerformVisualTest()
    EndFunction
EndProperty

String Property HorizontalAnchor
    String Function get()
        Return _horizontalAnchor
    EndFunction

    Function set(String a_val)
        ; If Config.AreTraceMessagesEnabled()
        ;     Config.Log("AproposWidgetManager.HorizontalAnchor.set(" + a_val + ")", True)
        ; EndIf
        _horizontalAnchor = a_val
		Int i = 0
		While i < Widgets.Length
			Widgets[i].Align = a_val
			i += 1
		EndWhile
        UpdateAllWidgetPositions()
        PerformVisualTest()
    EndFunction
EndProperty

String Property VerticalAnchor
    String Function get()
        Return _verticalAnchor
    EndFunction

    Function set(String a_val)
        ; If Config.AreTraceMessagesEnabled()
        ;     Config.Log("AproposWidgetManager.VerticalAnchor.set(" + a_val + ")", True)
        ; EndIf
        _verticalAnchor = a_val
        UpdateAllWidgetPositions()
        PerformVisualTest()
    EndFunction
EndProperty

Float Property Alpha
    Float Function get()
        Return _alpha
    EndFunction

    Function set(Float a_val)
        ; If Config.AreTraceMessagesEnabled()
        ;     Config.Log("AproposWidgetManager.Alpha.set(" + a_val + ")", True)
        ; EndIf
        _alpha = a_val
		Int i = 0
		While i < Widgets.Length
			Widgets[i].Alpha = a_val
			i += 1
		EndWhile
    EndFunction
EndProperty

Bool Property Ready
    Bool Function get()
    	Bool isReady = True
		Int i = 0
		While i < Widgets.Length
			isReady = isReady && Widgets[i].Ready
			i += 1
		EndWhile    	
		Return isReady
    EndFunction
EndProperty

Bool Property Initialized
    Bool Function get()
        Return _initialized
    EndFunction
EndProperty

Int[] Function GetOffsetPos(Int widgetIndex)
    Int[] offsetPos = New Int[2]

    AproposMessageWidget widget = Widgets[widgetIndex]
    
    If _horizontalAnchor == "Left"
        offsetPos[0] = 0
    ElseIf _horizontalAnchor == "Center"
    	offsetPos[0] = (-widget.GetDimensions()[0] / 2.0) As Int
    ElseIf _horizontalAnchor == "Right"
    	offsetPos[0] = -widget.GetDimensions()[0] As Int
    EndIf

    AproposMessageWidget firstWidget = Widgets[0]

    AproposMessageWidget previousWidget
    Float previousHeights

    If widgetIndex == 0
        previousWidget = firstWidget
        previousHeights = 0
    Else
        previousWidget = Widgets[widgetIndex - 1]
        Int i = 0
        While i < widgetIndex
            widget = Widgets[i]
            previousHeights += widget.GetDimensions()[1]
            i += 1
        EndWhile
    EndIf

    ;Config.Debug("PreviousHeights " + previousHeights + ", previousYpos " + previousWidget.Y)

    ; Assumes first widget has a unique size, and all others are sized vertically uniformly.

    If _verticalAnchor == "Top"
        offsetPos[1] = (previousHeights + (_widgetSpacing As Float)) As Int
        ;offsetPos[1] = ((widgetIndex As Float) * (previousHeight + (_widgetSpacing As Float))) As Int
    ElseIf _verticalAnchor == "Center"
        offsetPos[1] = (((widgetIndex - 5) As Float) * (previousHeights + (_widgetSpacing As Float))) As Int
    ElseIf _verticalAnchor == "Bottom"
        offsetPos[1] = ((widgetIndex As Float) * -(previousHeights + (_widgetSpacing As Float))) As Int
    EndIf

    ; If Config.AreDebugMessagesEnabled()
    ;     Config.Log("AproposWidgetManager.GetOffsetPos(" + widgetIndex + ") = (" + offsetPos[0] + "," + offsetPos[1] + ")")
    ; EndIf

    Return offsetPos
EndFunction

Function UpdatePosition(Int widgetIndex)
    Int[] refPos = GetReferencePos()
    Int[] offset = GetOffsetPos(widgetIndex)
    AproposMessageWidget widget = Widgets[widgetIndex]
    Int calc1 = refPos[0] + offset[0]
	Int calc2 = refPos[1] + offset[1]
    widget.MoveTo(calc1, calc2)
EndFunction

Function UpdateAllWidgetPositions()
	Int[] refPos = GetReferencePos()
    Int[] offset
    Int calc1
    Int calc2
	Int i = 0
    ; If Config.AreDebugMessagesEnabled()
    ;     Config.Log("AproposWidgetManager.UpdateAllWidgetPositions()")
    ; EndIf        

    While i < Widgets.Length
        offset = GetOffsetPos(i)
    	AproposMessageWidget widget = Widgets[i]
    	calc1 = refPos[0] + offset[0]
    	calc2 = refPos[1] + offset[1]
        ; If Config.AreDebugMessagesEnabled()
        ;     Config.Log("... widget[" + i + "].MoveTo(" + calc1 +" ," + calc2 + ")")
        ; EndIf        
    	widget.MoveTo(calc1, calc2)
    	i += 1
    EndWhile
EndFunction

Int[] Function GetReferencePos()
    Int[] refPos = New Int[2]
    
    If (_horizontalAnchor == "Left")
        refPos[0] = _horizontalOffset
    ElseIf (_horizontalAnchor == "Center")
        refPos[0] = 640 + _horizontalOffset
    ElseIf (_horizontalAnchor == "Right")
        refPos[0] = 1280 + _horizontalOffset
    EndIf
    
    If (_verticalAnchor == "Top")
        refPos[1] = _verticalOffset
    ElseIf (_verticalAnchor == "Center")
        refPos[1] = 360 + _verticalOffset
    ElseIf (_verticalAnchor == "Bottom")
        refPos[1] = 720 + _verticalOffset
    EndIf

    ; If Config.AreDebugMessagesEnabled()
    ;     Config.Log("AproposWidgetManager.GetReferencePos() = (" + refPos[0] + "," + refPos[1] + ")")
    ; EndIf
    
    Return refPos
EndFunction

Int Property    DefaultFontSize         = 16     AutoReadOnly
Int Property    DefaultPlayerFontSize   = 20     AutoReadOnly
String Property DefaultHorizontalAnchor = "Left" AutoReadOnly
String Property DefaultVerticalAnchor   = "Top"  AutoReadOnly
Int Property    DefaultHorizontalOffset = 5      AutoReadOnly
Int Property    DefaultVerticalOffset   = 134    AutoReadOnly
Int Property    DefaultWidgetSpacing    = 0      AutoReadOnly
Float Property  DefaultSecondsToDisplay = 7.0    AutoReadOnly
